# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device Power Default Capability."""
import time
from typing import Any, Callable, Dict

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import device_power_base
from gazoo_device.utility import deprecation_utils
import immutabledict

logger = gdm_logger.get_logger()

_HUB_TYPE_PROPERTY = "device_power_hub_type"
_HUB_TYPE_CAMBRIONIX = "cambrionix"
_HUB_TYPE_POWERSWITCH = "powerswitch"
_HUB_TYPE_UNIFI_SWITCH = "unifi_switch"

SUPPORTED_HUB_TYPES = (
    _HUB_TYPE_CAMBRIONIX, _HUB_TYPE_POWERSWITCH, _HUB_TYPE_UNIFI_SWITCH,
)

_HUB_TYPE_PROPS = immutabledict.immutabledict({
    _HUB_TYPE_CAMBRIONIX: ("device_usb_hub_name", "device_usb_port"),
    _HUB_TYPE_POWERSWITCH: ("powerswitch_name", "powerswitch_port"),
    _HUB_TYPE_UNIFI_SWITCH: ("unifi_switch_name", "unifi_switch_port"),
})


class DevicePowerDefault(device_power_base.DevicePowerBase):
  """Base class for device_power."""

  def __init__(self,
               device_name: str,
               create_device_func: Callable[..., Any],
               default_hub_type: str,
               props: Dict[str, Any],
               usb_ports_discovered: bool,
               wait_for_bootup_complete_fn: Callable[[], None],
               switchboard_inst: Any,
               change_triggers_reboot: bool = False,
               usb_hub_name_prop: str = "device_usb_hub_name",
               usb_port_prop: str = "device_usb_port"):
    """Create an instance of the device_power capability.

    The power switch type can be changed by setting the 'device_power_hub_type'
    property. The valid hub types are 'cambrionix', 'powerswitch', and
    'unifi_switch'.

    Each hub type also requires a corresponding hub_name and hub_port property
    to be set.

    Cambrionix requires 'device_usb_hub_name' and 'device_usb_port' to be set.
      (These names can be changed by passing in the different property names)

    Powerswitch requires 'powerswitch_name' and 'powerswitch_port' to be set.

    Unifi_switch requires 'unifi_switch_name' and 'unifi_switch_port' to be set.

    Args:
      device_name: Name of the device this capability is attached to.
      create_device_func: A method to create device of hub_type.
      default_hub_type: Type of switch for power cycling.
      props: Dictionary of device props from configuration file.
      usb_ports_discovered: True if the USB ports are discovered by gdm detect.
      wait_for_bootup_complete_fn: A method that the capability can
        call to wait for a reboot to complete if triggered by a change.
      switchboard_inst: Instance of switchboard capability
      change_triggers_reboot: Set change_triggers_reboot to TRUE if
        changing the power mode for the device causes a reboot.
      usb_hub_name_prop: Name of the hub name property.
      usb_port_prop: Name of the hub port property.

    Raises:
      ValueError: If hub type is not a known type.
    """
    super().__init__(device_name=device_name)

    self._hub = None
    self._create_device_func = create_device_func
    self._usb_ports_discovered = usb_ports_discovered
    self._props = props
    hub_type = self._props["optional"].get(_HUB_TYPE_PROPERTY, default_hub_type)
    if hub_type not in SUPPORTED_HUB_TYPES:
      raise ValueError("Hub type {} is not supported. Valid types: {}".format(
          hub_type, SUPPORTED_HUB_TYPES))
    self._hub_type = hub_type
    self._hub_name_prop, self._port_prop = _HUB_TYPE_PROPS[self._hub_type]
    if self._hub_type == _HUB_TYPE_CAMBRIONIX:
      self._hub_name_prop = usb_hub_name_prop
      self._port_prop = usb_port_prop

    # Set where the hub/port propeties are in _props.
    if self._hub_type == _HUB_TYPE_CAMBRIONIX and self._usb_ports_discovered:
      self._dict_name = "persistent_identifiers"
    else:
      self._dict_name = "optional"

    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._switchboard = switchboard_inst
    self._change_triggers_reboot = change_triggers_reboot

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks that the capability is ready to use.

    Raises:
      CapabilityNotReadyError: If expected properties are not set or
        unable to create auxiliary device for power switching.
    """
    unset_props = []
    if self.hub_name is None:
      unset_props.append(self._hub_name_prop)
    if self.port_number is None:
      unset_props.append(self._port_prop)
    if unset_props:
      if self._hub_type == _HUB_TYPE_CAMBRIONIX and self._usb_ports_discovered:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm redetect {}'")
      else:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm set-prop {} <property> <value>'")
      msg = msg_format.format(self._hub_type, self._device_name)
      error_msg = "properties {} are unset. ".format(
          " and ".join(unset_props)) + msg
      raise errors.CapabilityNotReadyError(
          msg=error_msg, device_name=self._device_name)
    try:
      self._hub = self._create_device_func(self.hub_name)
    except errors.DeviceError as err:
      raise errors.CapabilityNotReadyError(
          msg=str(err), device_name=self._device_name)
    self._healthy = True

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def close(self):
    """Closes the hub device instance."""
    if self._hub:
      self._hub.close()
    super().close()

  @decorators.OptionalProperty
  def hub_name(self):
    """Name of the hub the device is attached to."""
    return self._props[self._dict_name].get(self._hub_name_prop)

  @decorators.OptionalProperty
  def hub_type(self):
    """Device type of the hub."""
    return self._hub_type

  @decorators.DynamicProperty
  def port_mode(self):
    """port mode."""
    if not self.healthy:
      self.health_check()
    return self._hub.switch_power.get_mode(self.port_number)

  @decorators.OptionalProperty
  def port_number(self):
    """Port number the device is attached to."""
    return self._props[self._dict_name].get(self._port_prop)

  @decorators.CapabilityLogDecorator(logger)
  def cycle(self, no_wait=False):
    """Power off then power on the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    self.off()
    time.sleep(2)  # Small delay before calling power_on
    self.on(no_wait=no_wait)

  @decorators.CapabilityLogDecorator(logger)
  def off(self, close_transports=True):
    """Turn off power to the device.

    Args:
      close_transports (bool): Whether to close transports when powering off.
    """
    if not self.healthy:
      self.health_check()
    if self._change_triggers_reboot:
      self._switchboard.add_log_note(
          f"GDM triggered reboot via {self.hub_type} power change.")
      self._hub.switch_power.power_off(self.port_number)
      self._wait_for_bootup_complete_fn()
    else:
      if close_transports:
        self._switchboard.close_all_transports()
      self._hub.switch_power.power_off(self.port_number)

  @decorators.CapabilityLogDecorator(logger)
  def on(self, no_wait=False):
    """Turn on power to the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    if self._change_triggers_reboot:
      self._switchboard.add_log_note(
          f"GDM triggered reboot via {self.hub_type} power change.")
    self._hub.switch_power.power_on(self.port_number)
    if not self._change_triggers_reboot:
      time.sleep(2)  # Small delay to give time for 'dev/tty' to populate
      self._switchboard.open_all_transports()

    if not no_wait:
      self._wait_for_bootup_complete_fn()


deprecation_utils.add_deprecated_attributes(DevicePowerDefault,
                                            [("power_off", "off", True),
                                             ("power_on", "on", True),
                                             ("power_cycle", "cycle", True)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the Pigweed RPC echo capability."""
from typing import Any, Callable
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_echo_base
from gazoo_device.switchboard.transports import pigweed_rpc_transport

logger = gdm_logger.get_logger()


class PwRPCEchoDefault(pwrpc_echo_base.PwRPCEchoBase):
  """Pigweed RPC echo capability."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any]):
    """Create an instance of the PwRPCEchoDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The Switchboard.call method.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call

  @decorators.CapabilityLogDecorator(logger)
  def echo(self, msg: str) -> str:
    """Sends a message to the echo endpoint and returns the response back.

    Args:
      msg: Echo message to send.

    Returns:
      Echo message.

    Raises:
      DeviceError: Raises when RPC ack value is not True.
    """
    ack, echo_msg = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.echo_rpc,
        method_args=(msg,))
    if not ack:
      raise errors.DeviceError(f"{self._device_name} echo failed.")
    return echo_msg



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reusable utility functions for executing methods in parallel.

This will create multiple processes to execute device operations on multiple
devices. Device
interactions will be logged to the provided logger. Errors can optionally be
raised if device
methods fail.

Example usage:
    sample_function_results = parallel_utils.parallel_process(
        "sample_function",
        self._sample_function,
        device_instances)

    for result in sample_function_results:
        do_something(result)
"""
import multiprocessing
import time

from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils
from six.moves import queue

logger_gdm = gdm_logger.get_logger()

TIMEOUT_PROCESS = 600.0
TIMEOUT_TERMINATE_PROCESS = 5
REQUIRED_PROPS = ["name", "DEVICE_TYPE"]


def parallel_process(action_name,
                     fcn,
                     devices,
                     logger=None,
                     parameter_dicts=None,
                     timeout=TIMEOUT_PROCESS):
  """Concurrently apply function to each device.

  TODO(b/152442892): Increase the flexibility of this function.

  Args:
      action_name (str): terse description of the function
      fcn (function): function to execute in parallel
      devices (list): list of device objects
      logger (logger): logger object that will be passed to fcn
      parameter_dicts (dict): of arguments to send to the fcn by device_type
      timeout (int): seconds before terminating all the parallel processes

  Returns:
      list: results from parallel functions if return_results flag is
      specified. this will be
          a list of either an exception raised while running the function or
          data added
          manually to the multiprocessing queue.

  Raises:
      RuntimeError: if any of the parallel functions raise error or timeouts
      and return_results flag is not true
      AttributeError: if provided devices are missing required props.
  """
  if not parameter_dicts:
    parameter_dicts = {}

  # verify list of devices was recieved
  if not isinstance(devices, list):
    raise RuntimeError(
        "Devices should be a list. Instead its a {}. Devices: {}".format(
            type(devices), devices))

  # verify devices have required properties
  for prop in REQUIRED_PROPS:
    if not all(hasattr(device, prop) for device in devices):
      raise AttributeError("Devices must have {} property. Devices: {}".format(
          prop, devices))

  device_names = [device.name for device in devices]
  logger_gdm.info("Executing {} concurrently for {}s on devices {}".format(
      action_name, timeout, ",".join(device_names)))

  # create queues to manage parallel processes
  return_queue = multiprocessing.Queue()
  error_queue = multiprocessing.Queue()
  processes = []

  # initiate new process for each device with provided arguments
  for device in devices:
    device_type = device.DEVICE_TYPE
    args = (fcn, return_queue, error_queue)
    kwargs = {"device": device}
    if logger:
      kwargs["logger"] = logger
    if device_type in parameter_dicts:
      kwargs["parameter_dict"] = parameter_dicts[device_type]
    process = multiprocessing.Process(
        target=_sub_process, args=args, kwargs=kwargs)
    processes.append(process)

  # run each process in parallel
  deadline = time.time() + timeout
  for process in processes:
    common_utils.run_before_fork()
    process.start()
    common_utils.run_after_fork_in_parent()

  for process in processes:
    remaining_timeout = max(0, deadline - time.time())  # ensure timeout >= 0
    process.join(timeout=remaining_timeout)
    if process.is_alive():
      process.terminate()
      process.join(timeout=TIMEOUT_TERMINATE_PROCESS)

  # handle results from queue
  errors = get_messages_from_queue(error_queue)
  if errors:
    raise RuntimeError(", ".join(errors))
  return get_messages_from_queue(return_queue)


def issue_devices_parallel(method_name,
                           devices,
                           parameter_dicts=None,
                           timeout=TIMEOUT_PROCESS):
  """Concurrently issue a device command to multiple devices.

  Args:
      method_name (str): device method to call in parallel.
      devices (list): list of device objects.
      parameter_dicts (dict): of arguments to send to the method by
        device_type.
      timeout (int): seconds before terminating all the parallel processes.

  Returns:
      list: results of parallel method calls.

  Note:
      The nested function _issue_devices_parallel will be the function
      executed in separate
      processes for each device.
  """

  def _issue_devices_parallel(device, parameter_dict=None):
    """Execute device method."""
    parameter_dict = parameter_dict or {}
    method = getattr(device, method_name)
    return method(**parameter_dict)

  return parallel_process(
      method_name,
      _issue_devices_parallel,
      devices,
      parameter_dicts=parameter_dicts,
      timeout=timeout)


def get_messages_from_queue(mp_queue, timeout=0.01):
  """Safely get all messages from a multiprocessing queue.

  Args:
      mp_queue (queue): a multiprocess Queue instance
      timeout (float): seconds to block other processes out from the queue

  Returns:
      list: List of messages or an empty list if there weren't any
  """
  msgs = []
  # According to the python docs https://docs.python.org/2/library/multiprocessing.html
  # after putting an object on an empty queue there may be an
  #  infinitesimal delay before the queue's empty() method returns False
  #
  # We've actually run into this (SPT-1354) so we'll first kick the
  # tires with a get() and then see if there are more using empty()
  try:
    msgs.append(mp_queue.get(True, timeout))
  except queue.Empty:
    pass
  else:
    while not mp_queue.empty():
      msgs.append(mp_queue.get_nowait())
  return msgs


def _sub_process(fcn, return_queue, error_queue, **kwargs):
  """Helper function for calling methods in parallel.

  Args:
      fcn (function): function to execute
      return_queue (queue): a multiprocess Queue instance for results
      error_queue (queue): a multiprocess Queue instance for exceptions
      **kwargs (dict): other keyworded arguments to pass to the function
  """
  try:
    result = fcn(**kwargs)
    if result is not None:
      return_queue.put_nowait(result)
  except Exception as err:
    error_queue.put_nowait(repr(err))



"""Methods for mobly controller.

Imported into init.

To use:
import gazoo_device
from mobly import base_test

class Test(base_test.TestCase):
  def set_up_class(self):
    self.devices = self.register_controller(gazoo_device)


  def setup_test(self):
    for device in self.devices:
      device.start_new_log(log_name_prefix=self.current_test_info.name)

  def teardown_test(self):
    for device in self.devices:
      device.make_device_ready()
"""
import logging
import os
from typing import Any, Dict, List, Sequence

from gazoo_device import custom_types
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device import manager

# Device properties for gazoo device analytics.
_DEVICE_PROPS = ("name", "device_type", "model", "platform", "serial_number",
                 "wifi_mac_address", "firmware_version", "firmware_branch",
                 "firmware_type", "alias", "communication_address",
                 "secondary_communication_address", "build_date",
                 "initial_code_name")
_LOGGER = gdm_logger.get_logger()
_MANAGER_INSTANCE = None


def _set_auxiliary_props(properties: Dict[str, Any], device_name: str):
  """Sets props on device instance to the values in the config.

  Label is translated to Alias and dimensions are left out.

  Args:
    properties: Dictionary of device properties and their values.
    device_name: Id of the device instance.
  """
  for prop_name, value in properties.items():
    if prop_name in ["id", "dimensions"]:
      continue
    elif prop_name == "label":
      _MANAGER_INSTANCE.set_prop(device_name, "alias", value)
    else:
      try:
        _MANAGER_INSTANCE.set_prop(device_name, prop_name, value)
      except errors.DeviceError:
        _LOGGER.warning("%s unsettable for %s", prop_name, device_name)


def create(configs: List[Dict[str, Any]]) -> List[custom_types.Device]:
  """Creates gazoo device instances and returns them."""
  # log_path is set by mobly on logging in base_test
  log_path_directory = getattr(logging, "log_path", "/tmp/logs")
  gdm_log_file = os.path.join(log_path_directory, "gdm.txt")
  global _MANAGER_INSTANCE
  _MANAGER_INSTANCE = manager.Manager(
      log_directory=log_path_directory,
      gdm_log_file=gdm_log_file,
      stdout_logging=False)
  devices = []
  for entry in configs:
    name = entry["id"]
    _set_auxiliary_props(entry, name)
    device = _MANAGER_INSTANCE.create_device(name)
    devices.append(device)
  return devices


def get_info(devices: Sequence[custom_types.Device]) -> List[Dict[str, Any]]:
  """Returns persistent info and firmware version for each device."""
  info = []
  for device in devices:
    props = {}
    for device_prop in _DEVICE_PROPS:
      try:
        props[device_prop] = device.get_property(device_prop, raise_error=True)
      except AttributeError:
        props[device_prop] = "Undefined"
      except errors.DeviceError as e:
        props[device_prop] = repr(e)
    info.append(props)
  return info


def destroy(devices: List[custom_types.Device]) -> None:
  """Closes all created devices and manager."""
  for device in devices:
    device.close()
  global _MANAGER_INSTANCE
  _MANAGER_INSTANCE.close()
  _MANAGER_INSTANCE = None



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global variables containing all extensions known to GDM.

Extension modules are registered by package_registrar.py at runtime.
Extension modules consist of:
- primary, auxiliary, and virtual device classes (controllers);
- capability interfaces and capability flavors (implementations);
- communication types;
- detection criteria for each communication type;
- metadata (extension package name and version);
- keys (such as SSH keys).

These values are intended for internal GDM usage only.
"""
auxiliary_devices = []  # List of device classes
# "capabilities" are derived from "capability_interfaces"
capabilities = {}  # Capability name -> capability interface name
capability_interfaces = {}  # Capability interface name -> interface class
capability_flavors = {}  # Capability flavor name -> capability flavor class
communication_types = {}  # Communication type name -> communication type class
# "detect_criteria" is a mapping of
# Communication type name -> {Query Key: Query function}
detect_criteria = {}
primary_devices = []  # List of device classes
virtual_devices = []  # List of device classes
# "package_info" is a mapping of Package name ->
# immutabledict({"version": Version, "key_download_function": function})
package_info = {}
keys = []  # List of data_types.KeyInfo instances
manager_cli_mixins = []  # List of class objects inheriting from FireManager


def get_registered_package_info() -> str:
  """Returns names and versions of all registered extension packages."""
  extension_versions = []
  for package_name, package_data in package_info.items():
    extension_versions.append(f"{package_name} {package_data['version']}")
  return ", ".join(extension_versions)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lighting_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='lighting_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16lighting_service.proto\x12\x08\x63hip.rpc\"\x0f\n\rLightingEmpty\"0\n\rLightingColor\x12\x0b\n\x03hue\x18\x01 \x01(\r\x12\x12\n\nsaturation\x18\x02 \x01(\r\"p\n\rLightingState\x12\n\n\x02on\x18\x01 \x01(\x08\x12\x12\n\x05level\x18\x02 \x01(\rH\x00\x88\x01\x01\x12+\n\x05\x63olor\x18\x03 \x01(\x0b\x32\x17.chip.rpc.LightingColorH\x01\x88\x01\x01\x42\x08\n\x06_levelB\x08\n\x06_color2\x80\x01\n\x08Lighting\x12\x39\n\x03Set\x12\x17.chip.rpc.LightingState\x1a\x17.chip.rpc.LightingEmpty\"\x00\x12\x39\n\x03Get\x12\x17.chip.rpc.LightingEmpty\x1a\x17.chip.rpc.LightingState\"\x00\x62\x06proto3'
)




_LIGHTINGEMPTY = _descriptor.Descriptor(
  name='LightingEmpty',
  full_name='chip.rpc.LightingEmpty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=36,
  serialized_end=51,
)


_LIGHTINGCOLOR = _descriptor.Descriptor(
  name='LightingColor',
  full_name='chip.rpc.LightingColor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='hue', full_name='chip.rpc.LightingColor.hue', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='saturation', full_name='chip.rpc.LightingColor.saturation', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=101,
)


_LIGHTINGSTATE = _descriptor.Descriptor(
  name='LightingState',
  full_name='chip.rpc.LightingState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='on', full_name='chip.rpc.LightingState.on', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='chip.rpc.LightingState.level', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color', full_name='chip.rpc.LightingState.color', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_level', full_name='chip.rpc.LightingState._level',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_color', full_name='chip.rpc.LightingState._color',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=103,
  serialized_end=215,
)

_LIGHTINGSTATE.fields_by_name['color'].message_type = _LIGHTINGCOLOR
_LIGHTINGSTATE.oneofs_by_name['_level'].fields.append(
  _LIGHTINGSTATE.fields_by_name['level'])
_LIGHTINGSTATE.fields_by_name['level'].containing_oneof = _LIGHTINGSTATE.oneofs_by_name['_level']
_LIGHTINGSTATE.oneofs_by_name['_color'].fields.append(
  _LIGHTINGSTATE.fields_by_name['color'])
_LIGHTINGSTATE.fields_by_name['color'].containing_oneof = _LIGHTINGSTATE.oneofs_by_name['_color']
DESCRIPTOR.message_types_by_name['LightingEmpty'] = _LIGHTINGEMPTY
DESCRIPTOR.message_types_by_name['LightingColor'] = _LIGHTINGCOLOR
DESCRIPTOR.message_types_by_name['LightingState'] = _LIGHTINGSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LightingEmpty = _reflection.GeneratedProtocolMessageType('LightingEmpty', (_message.Message,), {
  'DESCRIPTOR' : _LIGHTINGEMPTY,
  '__module__' : 'lighting_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.LightingEmpty)
  })
_sym_db.RegisterMessage(LightingEmpty)

LightingColor = _reflection.GeneratedProtocolMessageType('LightingColor', (_message.Message,), {
  'DESCRIPTOR' : _LIGHTINGCOLOR,
  '__module__' : 'lighting_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.LightingColor)
  })
_sym_db.RegisterMessage(LightingColor)

LightingState = _reflection.GeneratedProtocolMessageType('LightingState', (_message.Message,), {
  'DESCRIPTOR' : _LIGHTINGSTATE,
  '__module__' : 'lighting_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.LightingState)
  })
_sym_db.RegisterMessage(LightingState)



_LIGHTING = _descriptor.ServiceDescriptor(
  name='Lighting',
  full_name='chip.rpc.Lighting',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=218,
  serialized_end=346,
  methods=[
  _descriptor.MethodDescriptor(
    name='Set',
    full_name='chip.rpc.Lighting.Set',
    index=0,
    containing_service=None,
    input_type=_LIGHTINGSTATE,
    output_type=_LIGHTINGEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='chip.rpc.Lighting.Get',
    index=1,
    containing_service=None,
    input_type=_LIGHTINGEMPTY,
    output_type=_LIGHTINGSTATE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LIGHTING)

DESCRIPTOR.services_by_name['Lighting'] = _LIGHTING

# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DataFramer classes split incoming raw data into complete lines.

The NewlineFramer (default) splits incoming raw data into partial and complete
lines at newline character boundaries.

The InterwovenLogFramer finds and extracts log lines interwoven with responses
using the log line regex provided.

Notes:
A complete line returned by all DataFramer based classes should include the
trailing newline character as is.

Partial lines are typically those raw bytes that remain which don't have a
trailing newline character.
"""
import abc
import re


class DataFramer(abc.ABC):
  """Abstract class for transforming raw device input into discreet lines."""

  @abc.abstractmethod
  def get_lines(self, raw_data, begin=0, end=None):
    """Yields each line from the raw_data provided from begin to end.

    Args:
        raw_data (str): to split into framed lines
        begin (int): index to the first character to start looking for newlines
        end (int): index to the last character to look for newline characters

    Yields:
        str: A partial or complete framed line found in the raw_data
        provided.

    Note:
        The begin argument allows for a slight optimization by skipping
        past previous partial line in search of the next newline character.
    """


def _validate_end_and_begin_are_valid(begin, end, string_len):
  """Validate begin and end are in the right range and are integers.

  Args:
      begin (int): character position to start searching for newlines from.
      end (int): end character position to stop searching for newlines from.
      string_len (int): length of the string to search for newlines.

  Raises:
     ValueError: if begin or end are not valid intput.
  """
  if not isinstance(begin, int):
    raise ValueError("Expected begin to be an integer found {}".format(
        type(begin)))
  elif begin < 0 or begin > string_len:
    raise ValueError(
        "Expected begin value to be >= 0 and <= {}, found {}".format(
            string_len, begin))
  elif end is not None:
    if not isinstance(end, int):
      raise ValueError("Expected end to be an integer found {}".format(
          type(end)))
    elif end < 0 or end > string_len:
      raise ValueError(
          "Expected end value to be >= 0 and <= {}, found {}".format(
              string_len, end))
    elif begin > end:
      raise ValueError("Expected end value {} to be <= begin value {}".format(
          end, begin))


def split_newlines_only(string,
                        begin=0,
                        end=None,
                        keepends=True,
                        cleanends=False):
  """Generates substrings of string up to and including newline characters only.

  Args:
      string (str): to extract substrings from.
      begin (int): character to start searching for newlines from, not the
        first character of the first substring returned.
      end (int): end character to stop searching for newlines from
      keepends (bool): flag indicating newline character should be returned as
        part of substring returned.
      cleanends (bool): flag indicating trailing line return and line feeds
        should be stripped off and replaced with a line feed.

  Raises:
      ValueError: if begin or end < 0 or > than length of the string

  Yields:
      str: Generates substrings of string from 0 to len(string) for each
           newline character found and optionally including the newline
           character.

  Note:
      The begin argument allows for skipping characters that you know won't
      contain newline characters. The first substring returned will always be
      from the first character of the string to the first newline character
      found or the end of the string if no newline characters exist.
      If the end argument is None then the length of the string will be used
      instead. If the end argument is specified and is less than the length
      of the string then this function will yield all characters from end to
      the length of the string as the last substring.
  """
  string_len = len(string)
  _validate_end_and_begin_are_valid(begin, end, string_len)
  if end is None:
    end = string_len
  start_index = 0
  if cleanends:
    pre_line_return = begin
    # If begin > 0 might need to find pre_line_return spot in string
    for i in range(begin, 0, -1):
      if string[i] != "\r" and string[i] != "\n":
        pre_line_return = i + 1
        break
  end_offset = 1 if keepends else 0
  for i in range(begin, end):
    if cleanends and string[i] != "\r" and string[i] != "\n":
      pre_line_return = i + 1
    if string[i] == "\n":
      if cleanends:
        yield string[start_index:pre_line_return] + string[i:i + end_offset]
      else:
        yield string[start_index:i + end_offset]
      start_index = i + 1
      if start_index == end:
        break
  else:
    yield string[start_index:]


class InterwovenLogFramer(DataFramer):
  """Finds and extract out of order logs that are interwoven with responses."""

  def __init__(self,
               log_line_pattern,
               add_newline=False,
               keepends=True,
               cleanends=True):
    """Initializes LogFramer with the log line pattern provided.

    Args:
        log_line_pattern (str): to use to match and extract log lines from
          response lines.
        add_newline (bool): add a newline character to the extracted log
          line
        keepends (bool): flag indicating that trailing LF should be returned
        cleanends (bool): flag indicating that trailing CR and LF should be
          replaced with LF only.

    Note: The log_line_pattern should contain a single regular expression
      group that contains the complete log line, including the newline
      character (or use add_newline to add one), to be extracted from the
      response line.  The cleanends flag will strip off all CR and LF
      characters at the end of the line. If keepends is True the LF
      character will be returned. This is done before the log_line_pattern
      logic is performed so that add_newline only adds a LF character to the
      log_line string matched if enabled.
    """
    self.add_newline = add_newline
    self.keepends = keepends
    self.cleanends = cleanends
    self.log_line_re = re.compile(log_line_pattern)

  def get_lines(self, raw_data, begin=0, end=None):
    """Yields each deframed line from the raw_data provided from begin to end.

    Args:
        raw_data (str): to split into framed lines
        begin (int): index to the first character to start looking for
          newlines
        end (int): index to the last character to look for newline
          characters

    Yields:
        str: A partial or complete framed line found in the raw_data
        provided.

    Note:
        The begin argument allows for a slight optimization by skipping
        past previous partial line in search of the next newline character.
    """
    for raw_line in split_newlines_only(
        raw_data,
        begin=begin,
        end=end,
        keepends=self.keepends,
        cleanends=self.cleanends):
      if raw_line[-1] != "\n":
        yield raw_line
      else:
        match = self.log_line_re.search(raw_line)
        if match:
          line1 = match.group(1)
          if self.add_newline:
            line1 += u"\n"
          yield line1
          line2 = raw_line[:match.start()] + raw_line[match.end():]
          if line2:
            yield line2
        else:
          yield raw_line


class NewlineFramer(DataFramer):
  """Splits lines at each newline character.

  Note:
      This is the default DataFramer.
  """
  def __init__(self, keepends=True, cleanends=True):
    """Initializes NewlineFramer with the options specified.

    Args:
        keepends (bool): flag indicating that trailing LF should be returned
        cleanends (bool): flag indicating that trailing CR and LF should be
          replaced with LF only.

    Note: The cleanends flag will strip off all CR and LF characters at the
      end of the line. If keepends is True the LF character will be
      returned. No LF characters are added in the process.
    """
    self.keepends = keepends
    self.cleanends = cleanends

  def get_lines(self, raw_data, begin=0, end=None):
    """Returns each deframed line from the raw_data provided from begin to end.

    Args:
        raw_data (str): to split into framed lines
        begin (int): index to the first character to start looking for
          newlines
        end (int): index to the last character to look for newline
          characters

    Returns:
        generator: A generator function that can be used to split raw_data
                   into partial or complete framed lines on newline characters.

    Note:
        The begin argument allows for a slight optimization by skipping
        past previous partial line in search of the next newline character.
    """
    return split_newlines_only(
        raw_data,
        begin=begin,
        end=end,
        keepends=self.keepends,
        cleanends=self.cleanends)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the PwRPC (Pigweed RPC) button capability."""
from typing import Any, Callable, Dict
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_button_base
from gazoo_device.switchboard.transports import pigweed_rpc_transport

logger = gdm_logger.get_logger()


class PwRPCButtonDefault(pwrpc_button_base.PwRPCButtonBase):
  """Pigweed RPC button capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               expect_button_regexes: Dict[int, str],
               expect_timeout: int,
               switchboard_call_expect: Callable[..., Any]):
    """Creates an instance of the PwRPCButtonDefault capability.

    Args:
      device_name: Device name used for logging.
      expect_button_regexes: Dict of expected regexes for each button id.
      expect_timeout: Timeout (s) to wait for the expected regex.
      switchboard_call_expect: The switchboard.call_and_expect method.
    """
    super().__init__(device_name=device_name)
    self._expect_button_regexes = expect_button_regexes
    self._expect_timeout = expect_timeout
    self._switchboard_call_expect = switchboard_call_expect

  @decorators.CapabilityLogDecorator(logger)
  def push(self, button_id: int) -> None:
    """Push the button with the given id.

    Args:
      button_id: Button ID.

    Raises:
      DeviceError: When button_id is invalid or RPC ack value is not True.
    """
    if button_id not in self._expect_button_regexes:
      valid_ids = set(self._expect_button_regexes.keys())
      raise errors.DeviceError(f"Invalid button id {button_id}. Possible valid "
                               f"ids: {valid_ids}.")

    expect_regex = self._expect_button_regexes[button_id]
    _, (ack, _) = self._switchboard_call_expect(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        pattern_list=[expect_regex],
        timeout=self._expect_timeout,
        method_args=("Button", "Event"),
        method_kwargs={"idx": button_id, "pushed": True},
        raise_for_timeout=True)

    if not ack:
      raise errors.DeviceError(f"Device {self._device_name} button {button_id} "
                               f"push failed.")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The minimum base class for switchboard subprocesses."""
import contextlib
import multiprocessing
import os
import signal
import socket
import traceback

from gazoo_device import gdm_logger
from gazoo_device.utility import common_utils
import psutil
import six.moves.queue

logger = gdm_logger.get_logger()


def get_message(queue, timeout=None):
  """Returns next message from queue.

  Args:
      queue (Queue): to retrieve next message from
      timeout (float): time to wait in seconds for incoming message to arrive.

  Raises:
      ValueError: if queue provided is not a valid queue object

  Returns:
      object: The message data or None if no messages were available or
              retrieved within timeout specified.
  """
  if queue is None or not hasattr(queue, "get") or not hasattr(
      queue, "task_done"):
    raise ValueError("Invalid queue of {} provided".format(type(queue)))

  message = None
  try:
    message = queue.get(block=True, timeout=timeout)
    queue.task_done()
    # manager shutdown or queue empty
  except (IOError, six.moves.queue.Empty, ValueError, socket.error):
    pass
  return message


def put_message(queue, message, timeout=None):
  """Puts message into queue using the timeout specified.

  Args:
      queue (Queue): to add message to
      message (object): to write into the queue
      timeout (float): time to wait in seconds for outgoing message to be
        accepted into queue.

  Raises:
      ValueError: if queue provided is not valid or has no put method
      Queue.Full: if outgoing queue is full and timeout was provided and was
                  reached before message was sent.
  """
  if queue is None or not hasattr(queue, "put"):
    raise ValueError("Invalid queue of {} provided".format(type(queue)))

  try:
    queue.put(message, block=True, timeout=timeout)
  except (IOError, socket.error):  # manager shutdown
    pass


@contextlib.contextmanager
def _child_process_wrapper(parent_proc, process_name, device_name,
                           exception_queue):
  """Wrapper to ignore interrupts in child processes; the main process handles cleanup."""
  signal.signal(signal.SIGINT, signal.SIG_IGN)

  try:
    yield
  except EOFError:  # parent process has terminated
    pass
  except Exception:
    stack_trace = traceback.format_exc()
    try:
      exception_queue.put_nowait("Device {} raised exception in {}. "
                                 "{}".format(device_name, process_name,
                                             stack_trace))
      os.kill(parent_proc.pid, signal.SIGUSR1)
    except (IOError, OSError):  # queue or parent process doesn't exist anymore
      pass


def _parent_is_alive(parent_proc):
  status = False
  try:
    status = parent_proc.status() != psutil.STATUS_ZOMBIE
  except psutil.NoSuchProcess:  # parent process doesn't exist anymore
    pass
  return status


def _process_loop(cls, parent_proc):
  """Child process loop which handles start/stop events and exceptions."""
  with _child_process_wrapper(parent_proc, cls.process_name, cls.device_name,
                              cls._exception_queue):
    try:
      cls._start_event.set()
    except Exception as err:
      stack_trace = traceback.format_exc()
      logger.info("Device {} Process {} error {!r} "
                  "start event error. {}".format(cls.device_name,
                                                 cls.process_name, err,
                                                 stack_trace))
    running = cls._pre_run_hook()
    while running and _parent_is_alive(parent_proc):
      try:
        if cls._terminate_event.is_set():
          cls._terminate_event.clear()
          break
      except IOError:  # manager shutdown
        break
      running = cls._do_work()
  try:
    cls._stop_event.set()
  except IOError:  # manager shutdown
    pass
  cls._post_run_hook()


class SwitchboardProcess(object):
  """Simplifies creating Switchboard processes.

  Provides the following capabilities:
      * starting and verifying process start
      * stopping and verifying process stop
      * checking if process was started
      * checking if process is still running
      * providing pre and post process loop hooks
      * signaling main process when exceptions are raised in subprocess
      * receiving queue commands with optional data to be processed
      * Simplifying process loop definition for subclasses
  """

  def __init__(self,
               device_name,
               process_name,
               mp_manager,
               exception_queue,
               command_queue,
               valid_commands=None):
    """Initialize SwitchboardProcess with the arguments provided.

    Args:
        device_name (str): of the device for exception error messages
        process_name (str): to use for process name and exception error
          messages
        mp_manager (multiprocessing.Manager): object to use for creating
          Events
        exception_queue (Queue): to use for reporting exception traceback
          message from subprocess
        command_queue (Queue): to receive commands into
        valid_commands (Optional[Tuple[str, ...]]): Valid command strings.
    """
    self.device_name = device_name
    self.process_name = process_name
    self._command_queue = command_queue
    self._exception_queue = exception_queue
    self._start_event = mp_manager.Event()
    self._stop_event = mp_manager.Event()
    self._terminate_event = mp_manager.Event()
    self._valid_commands = valid_commands or ()

  def __del__(self):
    if self.is_started():
      self.stop()

  def start(self):
    """Starts process if process is not already running.

    Raises:
        RuntimeError: if called when process is already running

    Note:
        When calling this method you should use is_started() to check if
        process was previously started to prevent raising an error.
    """
    if not self.is_started():
      parent_proc = psutil.Process(os.getpid())
      self._start_event.clear()
      self._stop_event.clear()
      process = multiprocessing.Process(
          name=self.process_name,
          target=_process_loop,
          args=(self, parent_proc))
      common_utils.run_before_fork()
      process.start()
      common_utils.run_after_fork_in_parent()
      start_event_value = self._start_event.wait(timeout=5)
      if not start_event_value:
        raise RuntimeError("Device {} failed to start child process {}. "
                           "Start event was not set.".format(
                               self.device_name, self.process_name))
      self._process = process
    else:
      raise RuntimeError("Device {} failed to start child process {}. "
                         "Child process is already running.".format(
                             self.device_name, self.process_name))

  def is_started(self):
    """Returns True if process was started, False otherwise.

    Return:
        bool: True if process was started
    """
    return hasattr(self, "_process")

  def is_running(self):
    """Returns True if process is currently running, False otherwise.

    Return:
        bool: True if process was started and is currently running

    Note:
        Use this method to determine if a process is still running, since
        the process might choose to self exit.
    """
    if self.is_started():
      return self._process.is_alive()
    return False

  def send_command(self, command, data=None):
    """Sends command with optional data provided.

    Args:
        command (str): string to send
        data (object): optional data to include with command.

    Raises:
        ValueError: if command provided is not in the list of valid commands
    """
    if command not in self._valid_commands:
      raise ValueError("Device {} send command to child process {} failed. "
                       "{} is not a valid command in {!r}".format(
                           self.device_name, self.process_name, command,
                           self._valid_commands))
    put_message(self._command_queue, (command, data))

  def stop(self):
    """Stops process if process is running.

    Raises:
        RuntimeError: if called when process is not currently running
        OSError: if a system related exception occurs.

    Note:
        When calling this method you should use is_started() to check if
        process was previously started to prevent raising an error.
    """
    if self.is_started():
      if self._process.is_alive():
        try:
          self._terminate_event.set()
        except IOError:  # manager shutdown
          pass
      try:
        stop_event_value = self._stop_event.wait(timeout=5)
        if not stop_event_value:
          msg = ("Device {} failed to stop child process {}. "
                 "Stop event was not set.").format(self.device_name,
                                                   self.process_name)
          logger.error(msg)
          raise IOError(msg)
      except (IOError, ValueError):  # manager shutdown failed
        pass
      self._process.join(timeout=1)
      if self._process.is_alive():
        self._process.terminate()
      delattr(self, "_process")
    else:
      msg = ("Device {} failed to stop child process {}. Child process is not "
             "currently running.").format(self.device_name, self.process_name)
      logger.error(msg)
      raise RuntimeError(msg)

  def is_command_done(self):
    """Returns True if command queue is empty, false otherwise.

    Returns:
        bool: True if command queue is empty
    """
    return self._command_queue.empty()

  def _do_work(self):
    """Perform subprocess work and indicate if process should continue.

    Returns:
        bool: indicating process should continue running

    Note:
        This implementation just returns False.
    """
    return False

  def _post_run_hook(self):
    """Cleanup hook for derived classes after process terminates.

    Note:
        Try to avoid raising any exceptions from this method as any exception
        raised will not reach the main process.
    """
    pass

  def _pre_run_hook(self):
    """Setup hook for derived classes before child process run loop begins.

    Returns:
        bool: A boolean value indicating the process loop should proceed.

    Note:
        Returning None or False from this method will prevent the process loop
        from proceeding.
    """
    return True



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Switch power capability interface.

This class defines the required API all flavors of the switch_power capability.
"""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class SwitchPowerBase(capability_base.CapabilityBase):
  """Abstract base class defining the API for the switch_power capability."""

  @abc.abstractproperty
  def supported_modes(self):
    """Get auxiliary device supported power modes."""

  @abc.abstractmethod
  def power_on(self, port):
    """This command powers on the port specified.

    Args:
        port (int): Identifies which auxiliary device port to power on.
    """

  @abc.abstractmethod
  def power_off(self, port):
    """This command powers off the port specified.

    Args:
        port (int): Identifies which auxiliary device port to power off.
    """

  @abc.abstractmethod
  def set_mode(self, mode, port):
    """Sets the given auxiliary device port to the mode specified.

    Args:
        mode (str): auxiliary device mode to set.
        port (int): The port to set.

    Raises:
        DeviceError: invalid port, or mode.
    """

  @abc.abstractmethod
  def get_mode(self, port):
    """Gets the auxiliary device mode for the specified port.

    Args:
        port (int): Use this port to get the mode.

    Returns:
        str: auxiliary device port mode settings

    Raises:
        DeviceError: invalid port.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all auxiliary devices."""
import difflib
import functools
import inspect
import os
import re
import time
from typing import Any, List, Set, Type
import weakref

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import extensions
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device_base
from gazoo_device.capabilities.interfaces import capability_base
from gazoo_device.switchboard import log_process
from gazoo_device.utility import common_utils
from gazoo_device.utility import deprecation_utils

logger = gdm_logger.get_logger()

ERROR_PREFIX = "Exception_"
TIMEOUTS = {"CONNECTED": 3, "SHELL": 60}


def get_log_filename(log_directory, device_name, name_prefix=""):
  """Returns the full path of log filename using the information provided.

  Args:
      log_directory (path): to where the log file should be created.
      device_name (str): to use in the log filename
      name_prefix (str): string to prepend to the start of the log file.

  Returns:
      str: Path to log filename using the information provided.
  """

  log_timestamp = time.strftime("%Y%m%d-%H%M%S")
  if name_prefix:
    log_file_name = "{}-{}-{}.txt".format(name_prefix, device_name,
                                          log_timestamp)
  else:
    log_file_name = "{}-{}.txt".format(device_name, log_timestamp)
  return os.path.join(log_directory, log_file_name)


class AuxiliaryDevice(auxiliary_device_base.AuxiliaryDeviceBase):
  """Interface containing required GDM APIs for auxiliary devices."""
  _OWNER_EMAIL = ""  # override in child classes

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    """Initialize a device class with the serial port information provided.

    Args:
        manager(manager.Manager): Manager object for this device instance.
        device_config (dict): dict of two dicts, 'optional' device
          parameters and 'persistent' device parameters.
        log_file_name (str): file name in the log directory for device logs.
        log_directory (str): directory in which the controller will create
          the log file.
    """
    self._log_object_lifecycle_event("__init__")
    self._manager_weakref = weakref.ref(manager)

    # Create a dictionary to store "properties".  For now keep the
    # classification of "persistent" and "optional".
    self.props = {
        "persistent_identifiers": device_config["persistent"],
        "optional": device_config["options"]
    }

    self._regexes = {}
    self._commands = {}
    self._timeouts = TIMEOUTS.copy()
    # Initialize log files
    self.log_directory = log_directory
    if log_file_name:
      self._log_file_name = os.path.join(log_directory, log_file_name)
    else:
      log_name_prefix = device_config["log_name_prefix"]
      self._log_file_name = get_log_filename(
          log_directory, self.name, name_prefix=log_name_prefix)
    self.device_type = self.DEVICE_TYPE
    self._user_count = 1  # Successive create_device calls just increment the
    # user count and reuse the same instance.

  @decorators.OptionalProperty
  def alias(self):
    """Returns the user-defined device alias (string)."""
    return self.props["optional"]["alias"]

  @decorators.PersistentProperty
  def communication_address(self):
    """Returns the name of the main communication port.

    (e.g. ip address).

    Returns:
        str: path or address of main communication port.
    """
    return self.props["persistent_identifiers"].get("console_port_name")

  @decorators.PersistentProperty
  def commands(self):
    """Dictionary of commands issued to the device via shell."""
    return self._commands

  @decorators.DynamicProperty
  def connected(self):
    """Returns whether or not device is connected."""
    device_config = {"persistent": self.props["persistent_identifiers"]}
    return self.is_connected(device_config)

  @classmethod
  def get_dynamic_property_names(cls):
    """Returns a list of dynamic property names including capability ones."""
    return cls._get_property_names(decorators.DynamicProperty)

  @classmethod
  def get_persistent_property_names(cls):
    """Returns a list of persistent property names including capability ones."""
    names = cls._get_property_names(decorators.PersistentProperty)
    for property_value in config.CLASS_PROPERTY_TYPES:
      names += cls._get_property_names(property_value)
    return sorted(names)

  @classmethod
  def get_optional_property_names(cls):
    """Returns a list of optional property names including capability ones."""
    return cls._get_property_names(decorators.OptionalProperty)

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [self.check_device_connected]

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def increment_user_count(self):
    """Increments the active user count for this device instance."""
    self._user_count += 1

  @decorators.DynamicProperty
  def log_file_name(self) -> str:
    """Current device log file name in use.

    Returns:
      Path to current device log file name.

    When the device has been recently created it might be possible that the log
    file path does not yet exist, but will be created very soon. The caller is
    expected to check if the file path returned exists. The caller should refer
    to this property often because log rotation might cause the log path to
    change depending on the max_log_size value currently in use.
    """
    current_log_filename = self._log_file_name

    # Check if log file has rotated to next log filename
    next_log_filename = log_process.get_next_log_filename(current_log_filename)
    while os.path.exists(next_log_filename):
      current_log_filename = next_log_filename
      next_log_filename = log_process.get_next_log_filename(
          current_log_filename)
    return current_log_filename

  @decorators.PersistentProperty
  def regexes(self):
    """Regular expressions used to retrieve properties, events and states from device output.

    Returns:
        dict: mapping of name to regular expression.
    """
    return self._regexes

  @decorators.PersistentProperty
  def model(self):
    """Returns the device model."""
    return self.props["persistent_identifiers"]["model"]

  @decorators.PersistentProperty
  def name(self):
    return self.props["persistent_identifiers"]["name"]

  @decorators.PersistentProperty
  def serial_number(self):
    return self.props["persistent_identifiers"]["serial_number"]

  @decorators.PersistentProperty
  def timeouts(self):
    """Dictionary of default timeouts to use when expecting certain actions."""
    return self._timeouts

  @decorators.health_check
  def check_create_switchboard(self):
    """Checks switchboard can be successfully created."""
    self.switchboard.add_log_note("{} switchboard successfully started.".format(
        self.name))

  @decorators.LogDecorator(logger)
  def check_device_ready(self):
    """Check device readiness by executing a series of health check methods."""
    self._execute_health_check_methods(self.health_checks)

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def close(self, force: bool = False):
    """Releases all resources if there are no more instance users.

    If there is at least one other instance user, does not close the device and
    only decrements user count.

    Args:
      force: If True, ignore the active user count and close the device even if
        there are remaining users.
    """
    self._log_object_lifecycle_event("close")
    if force:
      self._user_count = 0
      self._close()
    else:
      # If user_count > 1, there are other users of this device instance, and we
      # can't clean up yet. If user_count == 0, cleanup already ran before.
      run_cleanup = self._user_count == 1
      if self._user_count > 0:
        self._user_count -= 1
      if run_cleanup:
        self._close()

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def _close(self):
    """Releases all resources."""
    self.reset_all_capabilities()

    if hasattr(self, "_manager_weakref"):
      manager_instance = self._manager_weakref()
      if manager_instance is not None:
        # pylint: disable=protected-access
        if self.name in manager_instance._open_devices:
          self._log_object_lifecycle_event("_close")
          del manager_instance._open_devices[self.name]
      del manager_instance

  def __del__(self):
    self._log_object_lifecycle_event("__del__")
    self.close(force=True)

  @decorators.health_check
  def check_device_connected(self):
    """Checks that device shows up as a connection on the host machine.

    Raises:
       DeviceNotConnectedError: if device is not connected.
    """
    device_config = {"persistent": self.props["persistent_identifiers"]}
    try:
      common_utils.retry(
          func=self.is_connected,
          func_args=(device_config,),
          is_successful=bool,
          timeout=self.timeouts["CONNECTED"])
    except errors.CommunicationTimeoutError:
      raise errors.DeviceNotConnectedError(
          self.name, msg="device not reachable from host machine.")

  def get_capability_classes(
      self, capability_name: str) -> List[Type[capability_base.CapabilityBase]]:
    """Returns possible capability classes (flavors) for the capability.

    Args:
        capability_name (str): name of the capability, such as "switchboard"
          or "file_transfer".

    Returns:
        list[type]: capability classes that can be used for the capability.

    Raises:
        DeviceError: if the capability is not supported by the device.

    Note:
        For a given device class, most capabilities are implemented by just 1
        flavor. In a few cases, there are several flavors. Which flavor is used
        depends on the firmware version and is determined at runtime.
    """
    if not self.has_capabilities([capability_name]):
      raise errors.DeviceError(
          "{} get_capability_classes failed. "
          "Capability {} is not supported by {}."
          .format(self.name, capability_name, self.device_type))
    capability_property = getattr(type(self), capability_name)
    return list(capability_property.capability_classes)

  def get_manager(self) -> Any:  # Returns a Manager instance.
    """Returns the Manager instance if it is still alive.

    Raises:
      RuntimeError: If the Manager instance is no longer alive.
    """
    manager = self._manager_weakref()
    if manager is None:
      raise RuntimeError(f"{self.name}'s Manager is no longer alive. "
                         "Has the device been closed?'")
    return manager

  def get_dynamic_properties(self):
    """Returns a dictionary of prop, value for each dynamic property."""
    names = self.get_dynamic_property_names()
    return self._get_properties(names)

  def get_persistent_properties(self):
    """Returns a dictionary of prop, value for each persistent property."""
    names = self.get_persistent_property_names()
    return self._get_properties(names)

  def get_property(self, name, raise_error=False):
    """Retrieves a property value (can be nested).

    Args:
        name (str): name of a single property to fetch.
        raise_error (bool): raise error if unable to retrieve property

    Returns:
        object: value of the specified dynamic property.

    Raises:
        DeviceError: property value is not a plain data object.
        AttributeError: property doesn't exist.
        Exception: property is not retrieveable for arbitrary reason.

    Note:
        Expects property value to be a plain data object.
        Returns a string when property doesn't exist or property raises an
        error.
        Able to process both device properties (firmware_version) and
        capability properties (usb_hub.device_port)
    """
    instance = self
    try:
      if "." in name:  # capability property
        instance = getattr(self, name.split(".")[0])
        name = name.split(".")[1]

      value = getattr(instance, name)
      if callable(value):
        raise errors.DeviceError("{}'s {} is a method".format(self.name, name))
      return value
    except AttributeError:
      if name in self.props["optional"]:
        return self.props["optional"][name]
      if raise_error:
        raise
      close_matches = difflib.get_close_matches(name, self.get_property_names())
      return "{} does not have a known property '{}'. Close matches: {}".format(
          self.name, name, " or ".join(close_matches))
    except Exception as err:
      if raise_error:
        raise
      error_type = type(err).__name__
      logger.info("{} for {}, exception: {}".format(error_type, name, str(err)))
      return ERROR_PREFIX + error_type

  def get_property_names(self):
    """Returns a list of all property names."""
    full_list = list(self.get_dynamic_property_names())
    full_list.extend(self.get_persistent_property_names())
    full_list.extend(self.get_optional_property_names())
    full_list.extend(self.props["optional"])
    return list(set(full_list))

  def get_optional_properties(self):
    """Returns a dictionary of prop, value for each optional property."""
    names = self.get_optional_property_names()
    return self._get_properties(names)

  @classmethod
  def get_supported_capabilities(cls):
    """Returns a list of names of capabilities supported by this device class."""
    # Deduplicate names: there may be several flavors which share the same interface
    capability_names = {
        capability_class.get_capability_name()
        for capability_class in cls.get_supported_capability_flavors()
    }
    return sorted(list(capability_names))

  @classmethod
  def get_supported_capability_flavors(
      cls) -> Set[Type[capability_base.CapabilityBase]]:
    """Returns all capability flavor classes supported by this device class."""
    capability_classes = [
        member.capability_classes
        for _, member in inspect.getmembers(cls)
        if isinstance(member, decorators.CapabilityProperty)
    ]
    return functools.reduce(set.union, capability_classes, set())

  @classmethod
  def has_capabilities(cls, capability_names):
    """Check whether this device class supports all of the given capabilities.

    Args:
        capability_names (list): names of capabilities to check for.

    Note:
        capability_names (list): list of capability names.
            Capability names are strings. They can be:
                - capability names ("file_transfer"),
                - capability interface names ("filetransferbase"),
                - capability flavor names ("filetransferscp").
            If an interface name or capability name is specified, the behavior
            is identical: any capability flavor which implements the given
            interface will match. If a flavor name is specified, only that
            capability flavor will match. Different kinds of capability names
            can be used together (["usb_hub", "filetransferscp"]).

    Returns:
        bool: True if all of the given capabilities are supported by this
        device class, False otherwise.

    Raises:
        DeviceError: invalid type of capability_names argument OR
                     one of the capability names provided isn't recognized
                     by GDM.
    """
    valid_capability_names_types = (list, tuple, set)
    if not isinstance(capability_names, valid_capability_names_types):
      raise errors.DeviceError(
          "Invalid type of capability_names. Expected one of: {}, found: {}."
          .format(valid_capability_names_types, type(capability_names)))
    if not all(isinstance(cap_name, str) for cap_name in capability_names):
      raise errors.DeviceError(
          "All capability names must be of string type. Found: {}.".format(
              capability_names))

    capabilities = []  # Interface or flavor classes
    for cap_name in capability_names:
      cap_name = cap_name.lower()
      if cap_name in extensions.capability_flavors:
        interface_or_flavor = extensions.capability_flavors[cap_name]
      elif cap_name in extensions.capability_interfaces:
        interface_or_flavor = extensions.capability_interfaces[cap_name]
      elif cap_name in extensions.capabilities:
        interface_name = extensions.capabilities[cap_name]
        interface_or_flavor = extensions.capability_interfaces[interface_name]
      else:
        msg = "\n".join([
            "Capability {} is not recognized.".format(cap_name),
            "Supported capability interfaces: {}".format(
                extensions.capability_interfaces.keys()),
            "Supported capability flavors: {}".format(
                extensions.capability_flavors.keys()),
            "Supported capabilities: {}".format(
                extensions.capabilities.keys())
        ])
        raise errors.DeviceError(msg)
      capabilities.append(interface_or_flavor)

    supported_capabilities = cls.get_supported_capability_flavors()
    for requested_capability in capabilities:
      if not any(
          issubclass(supported_capability, requested_capability)
          for supported_capability in supported_capabilities):
        return False
    return True

  def is_capability_initialized(self, capability_name: str) -> bool:
    """Returns whether the capability has been initialized.

    Args:
        capability_name (str): name of the capability, such as "switchboard"
          or "file_transfer".

    Returns:
        bool: True if capability has been initialized, False otherwise.
    """
    capability_class = self.get_capability_classes(capability_name)[0]
    return hasattr(self, self._get_private_capability_name(capability_class))

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def make_device_ready(self, setting: str = "on") -> None:
    """Checks device readiness and attempts recovery if allowed.

    If setting is 'off': does nothing.
    If setting is 'check_only': only checks readiness (recovery is skipped).
    If setting is 'on': checks readiness and attempts recovery
    self._RECOVERY_ATTEMPTS times.
    If setting is 'flash_build': same as 'on', but will attempt reflashing the
    device if it's supported and if all other recovery methods fail.

    Args:
      setting: 'on', 'off', 'check_only', or 'flash_build'.

    Raises:
      CheckDeviceReadyError: Re-raises the device readiness check error if
        unable to recover from it.
      DeviceError: If the recovery process raises an error.
    """
    if setting == "off":
      return

    recoverable_error = None
    unrecoverable_error = None
    for attempt in range(self._RECOVERY_ATTEMPTS):
      logger.info(f"{self.name} checking device readiness: attempt "
                  f"{attempt + 1} of {self._RECOVERY_ATTEMPTS}")
      try:
        self.check_device_ready()
      except errors.CheckDeviceReadyError as err:
        if setting == "check_only":
          logger.info(f"{self.name} make_device_ready setting is {setting!r}. "
                      "Skipping device recovery")
          raise
        # pylint: disable=unidiomatic-typecheck
        if type(recoverable_error) == type(err):
          logger.info(
              f"{self.name} readiness check raised the same error type "
              f"{type(err).__name__!r} after a recovery attempt. "
              "Assuming that the device is unable to recover. "
              f"Previous error: {recoverable_error!r}. New error: {err!r}")
          unrecoverable_error = err
          break
        recoverable_error = err
      else:
        if attempt != 0:
          logger.info(f"{self.name} successfully recovered to a ready state "
                      f"after {attempt + 1} recovery attempts")
        return

      logger.info(f"{self.name} attempting recovery: attempt {attempt + 1} "
                  f"of {self._RECOVERY_ATTEMPTS}")
      try:
        self.recover(recoverable_error)
      except errors.DeviceError as err:
        logger.info(
            f"{self.name} recovery attempt {attempt + 1} raised an error. "
            f"Assuming that the device is unable to recover. Error: {err!r}")
        unrecoverable_error = err
        break
    else:
      logger.info(f"{self.name} ran out of recovery attempts. "
                  "Executing a final device readiness check.")
      try:
        self.check_device_ready()
      except errors.CheckDeviceReadyError as err:
        unrecoverable_error = err
      else:
        logger.info(f"{self.name} successfully recovered to a ready state "
                    f"after {self._RECOVERY_ATTEMPTS} recovery attempts")
        return

    # If execution gets here, either recovery failed due to an error during
    # recover() or due to seeing the same CheckDeviceReadyError after a
    # successful recover() call, or recovery ran out of attempts.
    if setting != "flash_build" or not self.has_capabilities(["flash_build"]):
      if setting == "flash_build":
        logger.warning(f"{self.name} make_device_ready setting is {setting!r}, "
                       "but device does not support 'flash_build' capability")
      logger.info(f"{self.name} was not able to recover from "
                  f"{unrecoverable_error!r}")
      raise unrecoverable_error

    logger.info(f"{self.name} re-flashing device with the default build")
    self.flash_build.upgrade(forced_upgrade=True)
    logger.info(f"{self.name} checking device readiness after reflashing")
    self.check_device_ready()
    logger.info(f"{self.name} successfully recovered to a ready state")

  @decorators.PersistentProperty
  def owner(self) -> str:
    """Email of the owner (maintainer) of this device class."""
    return self._OWNER_EMAIL

  @decorators.LogDecorator(logger)
  def recover(self, error: errors.CheckDeviceReadyError) -> None:
    """Attempts to recover device based on the type of error specified.

    Args:
      error: A subclass of CheckDeviceReadyError that will be used to identify
        a possible recovery solution to use.

    Raises:
      CheckDeviceReadyError: If there are no recovery steps available for
        the error argument, it will be re-raised directly.
    """
    raise error

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def reset_all_capabilities(self):
    """Resets all capabilities which have been initialized by deleting them.

    Capabilities will be re-initialized on next use (when they're accessed).
    """
    for capability_name in self.get_supported_capabilities():
      self.reset_capability(capability_name)

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def reset_capability(self, capability_name: str) -> None:
    """Resets the capability if it's already initialized by deleting it.

    Calls the "close" method of the capability before deleting it.
    The capability will be re-initialized on next capability use (when it's
    accessed).

    Args:
        capability_name (str): name of the capability, such as "switchboard"
        or "file_transfer".
    """
    capability_class = self.get_capability_classes(capability_name)[0]
    capability_name = self._get_private_capability_name(capability_class)
    if hasattr(self, capability_name):
      getattr(self, capability_name).close()
      delattr(self, capability_name)

  @decorators.LogDecorator(logger, decorators.DEBUG)
  def set_property(self, prop, value):
    """Set an optional property.

    Args:
        prop (str): property name
        value (object): value of the property name.

    Raises:
        ValueError: if property is persistent or dynamic with no setter.
    """
    if prop in self.get_persistent_property_names():
      raise ValueError("{}'s {} is a persistent property and not settable. "
                       "Redetect device if it's wrong.".format(self.name, prop))
    if prop in self.get_dynamic_property_names():
      if not self._property_has_setter(prop):
        raise ValueError(
            "{}'s {} is a dynamic property which is not settable.".format(
                self.name, prop))
      instance = self
      if "." in prop:  # capability property
        instance = getattr(self, prop.split(".")[0])
        prop = prop.split(".")[1]
      setattr(instance, prop, value)
      return
    self.props["optional"][prop] = value

  def command_with_regex(self,
                         command,
                         regex,
                         command_fn,
                         regex_group=1,
                         raise_error=False,
                         tries=1,
                         **command_fn_kwargs):
    """Sends a command, searches for a regex in the response, and returns a match group.

    Args:
        command (str): command to issue.
        regex (str): regular expression with one or more capturing groups.
        command_fn (func): the function used to send the command.
        regex_group (int): number of regex group to return.
        raise_error (bool): whether or not to raise error if unable to find
          a match.
        tries (int): how many times to try executing the command before
          failing.
        **command_fn_kwargs (dict): keyword arguments for the command_fn
          function.

    Returns:
        str: value of the capturing group with index 'regex_group' in the match.

    Raises:
        DeviceError: if command execution fails OR
                     couldn't find the requested group in any of the
                     responses.
    """
    for _ in range(tries):
      try:
        response = command_fn(command, **command_fn_kwargs)
      except errors.DeviceError:
        continue

      match = re.search(regex, response, re.MULTILINE | re.DOTALL)
      if match:
        max_group = match.lastindex
        if max_group is None:
          max_group = 0

        if regex_group > max_group:
          logger.warning(
              "{}: requested group index ({}) exceeds index of last matched group ({}). "
              "Matched groups: {}, response: {!r}, regex: {!r}.".format(
                  self.name, regex_group, max_group, match.groups(), response,
                  regex))
        else:
          return str(match.group(regex_group))

    msg = "{} unable to retrieve {} from {} after {} tries".format(
        self.name, regex, command, tries)

    if raise_error:
      raise errors.DeviceError(msg)
    else:
      logger.warning(msg)
      return ""

  @decorators.health_check
  def check_ping_responsiveness(self):
    """Check if the auxiliary device responds to pings.

    Raises:
        DeviceNotResponsiveError: if no response to ping before the timeout.
    """
    try:
      common_utils.retry(
          self._ping,
          is_successful=bool,
          timeout=self._PING_TIMEOUT,
          reraise=True)
    except Exception as err:
      raise errors.DeviceNotResponsiveError(
          self.name,
          "failed to respond to ping",
          timeout=self._PING_TIMEOUT,
          details=str(err))

  def _execute_health_check_methods(self, health_checks):
    """Execute health checks on the device.

    Args:
        health_checks (list): list of methods to execute as health checks.

    Raises:
        CheckDeviceReadyError: if a health check fails.

    Note:
        Order of health check methods matter. Health check execution will
        stop at the first failing health check.
    """
    for step, health_check_method in enumerate(health_checks):
      method_name = health_check_method.__name__
      health_check_name = method_name.replace("_", " ").strip().capitalize()

      try:
        health_check_method()
      except errors.CheckDeviceReadyError as err:
        logger.info("{} health check {}/{} {!r} failed: {!r}.".format(
            self.name, step + 1, len(health_checks), health_check_name, err))
        raise

      logger.info("{} health check {}/{} succeeded: {}.".format(
          self.name, step + 1, len(health_checks), health_check_name))

  def lazy_init(self, capability_class, *args, **kwargs):
    """Provides a lazy instantiation mechanism for capabilities.

    The capability instance will not be created until it is accessed for the
    first time.
    Subsequent accesses will return the same capability instance.
    In other words, no more than one capability instance will be created.
    If a capability is not accessed, the capability instance is not created.

    Args:
        capability_class (class): capability class to instantiate.
        *args (tuple): positional args to the capability's __init__.
        **kwargs (dict): keyword arguments to the capability's __init__.

    Returns:
        CapabilityBase: initialized capability instance.
    """
    capability_name = self._get_private_capability_name(capability_class)
    if not hasattr(self, capability_name):
      capability_inst = capability_class(*args, **kwargs)
      setattr(self, capability_name, capability_inst)
    return getattr(self, capability_name)

  @classmethod
  def _get_private_capability_name(cls, capability_class):
    """Returns the name of the private attribute to use for a given capability class.

    Args:
        capability_class(class): capability class object.

    Returns:
        str: name of the private capability attribute to be used.
    """
    return "_{}".format(capability_class.get_capability_name())

  @classmethod
  def _get_property_names(cls, property_type):
    """Returns the property names for all public properties of the type."""
    classes = cls.get_supported_capability_flavors().copy()
    classes.add(cls)
    property_names = []
    for a_class in classes:
      # pytype: disable=attribute-error
      prefix = "" if a_class == cls else a_class.get_capability_name() + "."
      # pytype: enable=attribute-error
      for name, member in inspect.getmembers(a_class):
        if isinstance(member, property_type) and not name.startswith("_"):
          property_names.append("{}{}".format(prefix, name))
    return list(set(property_names))

  def _get_properties(self, property_names):
    """Returns a dictionary of prop, value for each property."""
    property_dict = {}
    for name in property_names:
      value = self.get_property(name)
      if isinstance(value, str) and "does not have a known property" in value:
        continue  # property not supported in current flavor
      property_dict[name] = value

    return property_dict

  def _get_switchboard_if_initialized(self):
    """Returns self.switchboard if it's initialized, None otherwise."""
    if self.is_capability_initialized("switchboard"):
      return self.switchboard  # pytype: disable=attribute-error
    return None

  def _list_properties_dynamic_auxiliary_device(self):
    """Returns a set of dynamic properties.

    Returns:
        set: dynamic properties that can be queried during get-prop CLI/API
        operation.
    """
    return set(["connected"])

  def _log_object_lifecycle_event(self, method_name):
    """Logs a message about a lifecycle event of a python object.

    Args:
        method_name (str): name of the method called on the object.
    """
    logger.debug("{} called on {} (id = {}) in process {}.".format(
        method_name, self, id(self), os.getpid()))

  def _property_has_setter(self, property_name):
    """Checks whether a property has a setter.

    Args:
        property_name (str): name of the property. If property is a
          capability property then the string should be
          <capability_name>.<property_name>

    Raises:
        AttributeError: if property doesn't exist

    Returns:
        bool: True if property has a setter, else False
    """
    target_class = type(self)
    # Get the class that holds the property
    if "." in property_name:
      capability_name = property_name.split(".")[0]
      capability_instance = getattr(self, capability_name)
      target_class = type(capability_instance)
      property_name = property_name.split(".")[1]
    # Check whether the property has a setter
    attr = getattr(target_class, property_name)
    return attr.fset is not None


deprecation_utils.add_deprecated_attributes(AuxiliaryDevice, [
    ("console_port_name", "communication_address", False),
])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TransportBase is the standard GDM device transport interface.

Device transports perform device communications. Device transports are used by
TransportProcess instances.
"""
import abc
import copy
from typing import Any, Dict, List, Optional, Union

from gazoo_device.switchboard import transport_properties


class TransportBase(abc.ABC):
  """Interface for transport classes."""

  def __init__(self,
               auto_reopen: bool = False,
               open_on_start: bool = True) -> None:
    """Initializes the transport interface.

    Args:
      auto_reopen: Flag indicating transport should be reopened if
        unexpectedly closed.
      open_on_start: Flag indicating transport should be open on
        TransportProcess start.
    """
    if not hasattr(self, "_properties"):
      self._properties = {}
    self._properties.update({
        transport_properties.AUTO_REOPEN: auto_reopen,
        transport_properties.OPEN_ON_START: open_on_start
    })

  @abc.abstractmethod
  def is_open(self) -> bool:
    """Returns True if the transport is open.

    Returns:
        bool: True if transport is open, False otherwise.
    """

  @abc.abstractmethod
  def _open(self) -> None:
    """Opens the transport using properties provided during __init__.

    Override in derived classes.
    """

  @abc.abstractmethod
  def _close(self) -> None:
    """Closes the transport. Override in derived classes."""

  @abc.abstractmethod
  def _read(self,
            size: int = 1,
            timeout: Optional[float] = None) -> Union[bytes, str]:
    """Returns bytes read up to size within the specified timeout.

    Args:
      size: Maximum number of bytes to read within timeout seconds.
      timeout: Maximum seconds to wait to read bytes or indefinitely if None.

    Returns:
      Bytes or unicode read from the transport.
      Empty string if no bytes were read or the timeout was reached.
    """

  @abc.abstractmethod
  def _write(self, data: str, timeout: Optional[float] = None) -> int:
    """Writes the data provided within timeout seconds.

    Args:
      data: Bytes to be written out within timeout seconds.
      timeout: Maximum seconds to wait to write bytes or indefinitely if None.

    Returns:
      Number of bytes written.
    """

  def __del__(self) -> None:
    """Closes the transport on garbage collection."""
    self.close()

  def get_property(self, key: str, value: Optional[Any] = None) -> Any:
    """Returns property matching key specified or value if not set.

    Args:
      key: Property to retrieve.
      value: Default value to return if the property doesn't exist.

    Returns:
      The property value for key specified or value if key doesn't exist.
    """
    return self._properties.get(key, value)

  def get_all_properties(self) -> Dict[str, Any]:
    """Returns dictionary of property key-value pairs.

    Returns:
      A deep copy of all properties.
    """
    return copy.deepcopy(self._properties)

  def get_property_list(self) -> List[str]:
    """Returns a list of property keys that exist and can be set/retrieved.

    Returns:
      Property keys that can be set or retrieved.
    """
    return list(self._properties.keys())

  def set_property(self, key: str, value: Any) -> None:
    """Sets the property specified to the value provided.

    Args:
      key: Property to set.
      value: Value to set the property to.

    Raises:
      KeyError: If property key doesn't exist.
    """
    if key not in self._properties:
      raise KeyError("Property {} doesn't exist for {}.".format(
          key, self.__class__.__name__))
    self._properties[key] = value

  def open(self):
    """Opens the transport if not open already."""
    if self.is_open():
      return None

    self._open()

  def close(self):
    """Closes the transport if it's open."""
    if not self.is_open():
      return None

    self._close()

  def read(self,
           size: int = 1,
           timeout: Optional[float] = None) -> Optional[Union[bytes, str]]:
    """Returns bytes read up to size within the specified timeout.

    Args:
      size: Maximum number of bytes to read within timeout seconds.
      timeout: Maximum seconds to wait to read bytes or indefinitely if None.

    Returns:
      Bytes or unicode read from the transport.
      Empty string if no bytes were read or the timeout was reached.
      None if the transport isn't open.
    """
    if not self.is_open():
      return None

    return self._read(size, timeout)

  def write(self, data: str, timeout: Optional[float] = None) -> Optional[int]:
    """Writes the data provided within timeout seconds if transport is open.

    Args:
      data: Bytes to be written out within timeout seconds.
      timeout: Maximum seconds to wait to write bytes or indefinitely if None.

    Returns:
      Number of bytes written if the transport is open.
      None if the transport isn't open.
    """
    if not self.is_open():
      return None

    return self._write(data, timeout)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for interaction with PwRPC (Pigweed RPC)."""
import enum
import os
from typing import Callable

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.protos import button_service_pb2
from gazoo_device.protos import device_service_pb2
from gazoo_device.protos import echo_service_pb2
from gazoo_device.protos import lighting_service_pb2
from gazoo_device.protos import locking_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport

_PWRPC_PROTOS = (button_service_pb2,
                 device_service_pb2,
                 echo_service_pb2,
                 lighting_service_pb2,
                 locking_service_pb2)


class PigweedAppType(enum.Enum):
  NON_PIGWEED = "nonpigweed"
  LIGHTING = "lighting"
  ECHO = "echo"
  LOCKING = "locking"

logger = gdm_logger.get_logger()


# Element format in _PIGWEED_APP_ENDPOINTS container:
# tuple of 3 elements: (method_args, method_kwargs, application_type)
# where method_args, method_kwargs are the input arguments
# to the PigweedRPCTransport.rpc,
# application_type is the Pigweed device type in string.
_PIGWEED_APP_ENDPOINTS = (
    (("Lighting", "Get"), {}, PigweedAppType.LIGHTING),
    (("msg",), {}, PigweedAppType.ECHO),
    (("Locking", "Get"), {}, PigweedAppType.LOCKING)
)


def get_application_type(
    address: str,
    log_path: str,
    create_switchboard_func: Callable[..., switchboard_base.SwitchboardBase]
) -> str:
  """Returns Pigweed application type of the device.

  Args:
    address: Device serial address.
    log_path: Device log path.
    create_switchboard_func: Method to create the switchboard.

  Returns:
    Pigweed application type.
  """
  switchboard = create_switchboard_func(communication_address=address,
                                        communication_type="PigweedSerialComms",
                                        device_name=os.path.basename(log_path),
                                        log_path=log_path,
                                        protobufs=_PWRPC_PROTOS)
  try:
    for method_args, method_kwargs, app_type in _PIGWEED_APP_ENDPOINTS:
      try:
        if app_type == PigweedAppType.ECHO:
          method = pigweed_rpc_transport.PigweedRPCTransport.echo_rpc
        else:
          method = pigweed_rpc_transport.PigweedRPCTransport.rpc
        switchboard.call(method=method,
                         method_args=method_args,
                         method_kwargs=method_kwargs)
        return app_type.value
      except errors.DeviceError:
        logger.info(f"Device {address} is not a Pigweed {app_type} device.")
  finally:
    switchboard.close()
  return PigweedAppType.NON_PIGWEED.value



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: button_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='button_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x62utton_service.proto\x12\x08\x63hip.rpc\"\r\n\x0b\x42uttonEmpty\"*\n\x0b\x42uttonEvent\x12\x0b\n\x03idx\x18\x01 \x01(\r\x12\x0e\n\x06pushed\x18\x02 \x01(\x08\x32\x41\n\x06\x42utton\x12\x37\n\x05\x45vent\x12\x15.chip.rpc.ButtonEvent\x1a\x15.chip.rpc.ButtonEmpty\"\x00\x62\x06proto3'
)




_BUTTONEMPTY = _descriptor.Descriptor(
  name='ButtonEmpty',
  full_name='chip.rpc.ButtonEmpty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=47,
)


_BUTTONEVENT = _descriptor.Descriptor(
  name='ButtonEvent',
  full_name='chip.rpc.ButtonEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='idx', full_name='chip.rpc.ButtonEvent.idx', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pushed', full_name='chip.rpc.ButtonEvent.pushed', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=91,
)

DESCRIPTOR.message_types_by_name['ButtonEmpty'] = _BUTTONEMPTY
DESCRIPTOR.message_types_by_name['ButtonEvent'] = _BUTTONEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ButtonEmpty = _reflection.GeneratedProtocolMessageType('ButtonEmpty', (_message.Message,), {
  'DESCRIPTOR' : _BUTTONEMPTY,
  '__module__' : 'button_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.ButtonEmpty)
  })
_sym_db.RegisterMessage(ButtonEmpty)

ButtonEvent = _reflection.GeneratedProtocolMessageType('ButtonEvent', (_message.Message,), {
  'DESCRIPTOR' : _BUTTONEVENT,
  '__module__' : 'button_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.ButtonEvent)
  })
_sym_db.RegisterMessage(ButtonEvent)



_BUTTON = _descriptor.ServiceDescriptor(
  name='Button',
  full_name='chip.rpc.Button',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=93,
  serialized_end=158,
  methods=[
  _descriptor.MethodDescriptor(
    name='Event',
    full_name='chip.rpc.Button.Event',
    index=0,
    containing_service=None,
    input_type=_BUTTONEVENT,
    output_type=_BUTTONEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BUTTON)

DESCRIPTOR.services_by_name['Button'] = _BUTTON

# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines an ssh transport interface by extending the process_transport.py."""
from typing import Optional

from gazoo_device import data_types
from gazoo_device.switchboard.transports import process_transport
from gazoo_device.utility import host_utils


class SSHTransport(process_transport.ProcessTransport):
  """Transport for communicating over SSH with a device."""

  def __init__(self,
               comms_address: str,
               args: str = host_utils.DEFAULT_SSH_OPTIONS,
               key_info: Optional[data_types.KeyInfo] = None,
               log_cmd: str = "",
               auto_reopen: bool = True,
               open_on_start: bool = True,
               username: str = "root"):
    """Initializes an SSHTransport object with the given properties.

    Args:
      comms_address: IP address of the device.
      args: Extra command line args for the SSH command.
      key_info: SSH key to use. If None, don't use an SSH key.
      log_cmd: Command to execute over SSH, such as "logcat".
      auto_reopen: Flag indicating whether the transport should be automatically
        reopened if unexpectedly closed.
      open_on_start: Flag indicating whether the transport should be open on
        TransportProcess start.
      username: Username to log in as.
    """
    self.comms_address = comms_address
    args = host_utils.generate_ssh_args(
        comms_address,
        log_cmd,
        username,
        options=args,
        key_info=key_info)
    super().__init__(
        command="ssh",
        args=args,
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)

  def _is_ready_to_open(self):
    return host_utils.is_pingable(self.comms_address)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: device_service.proto
# pylint: skip-file
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='device_service.proto',
  package='chip.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14\x64\x65vice_service.proto\x12\x08\x63hip.rpc\"\r\n\x0b\x44\x65viceEmpty\"d\n\nDeviceInfo\x12\x11\n\tvendor_id\x18\x01 \x01(\r\x12\x12\n\nproduct_id\x18\x02 \x01(\r\x12\x18\n\x10software_version\x18\x03 \x01(\r\x12\x15\n\rserial_number\x18\x04 \x01(\t2\x80\x02\n\x06\x44\x65vice\x12>\n\x0c\x46\x61\x63toryReset\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12\x38\n\x06Reboot\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12<\n\nTriggerOta\x12\x15.chip.rpc.DeviceEmpty\x1a\x15.chip.rpc.DeviceEmpty\"\x00\x12>\n\rGetDeviceInfo\x12\x15.chip.rpc.DeviceEmpty\x1a\x14.chip.rpc.DeviceInfo\"\x00\x62\x06proto3'
)




_DEVICEEMPTY = _descriptor.Descriptor(
  name='DeviceEmpty',
  full_name='chip.rpc.DeviceEmpty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=47,
)


_DEVICEINFO = _descriptor.Descriptor(
  name='DeviceInfo',
  full_name='chip.rpc.DeviceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vendor_id', full_name='chip.rpc.DeviceInfo.vendor_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='product_id', full_name='chip.rpc.DeviceInfo.product_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_version', full_name='chip.rpc.DeviceInfo.software_version', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serial_number', full_name='chip.rpc.DeviceInfo.serial_number', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=149,
)

DESCRIPTOR.message_types_by_name['DeviceEmpty'] = _DEVICEEMPTY
DESCRIPTOR.message_types_by_name['DeviceInfo'] = _DEVICEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeviceEmpty = _reflection.GeneratedProtocolMessageType('DeviceEmpty', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEEMPTY,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceEmpty)
  })
_sym_db.RegisterMessage(DeviceEmpty)

DeviceInfo = _reflection.GeneratedProtocolMessageType('DeviceInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEINFO,
  '__module__' : 'device_service_pb2'
  # @@protoc_insertion_point(class_scope:chip.rpc.DeviceInfo)
  })
_sym_db.RegisterMessage(DeviceInfo)



_DEVICE = _descriptor.ServiceDescriptor(
  name='Device',
  full_name='chip.rpc.Device',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=152,
  serialized_end=408,
  methods=[
  _descriptor.MethodDescriptor(
    name='FactoryReset',
    full_name='chip.rpc.Device.FactoryReset',
    index=0,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Reboot',
    full_name='chip.rpc.Device.Reboot',
    index=1,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TriggerOta',
    full_name='chip.rpc.Device.TriggerOta',
    index=2,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEEMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDeviceInfo',
    full_name='chip.rpc.Device.GetDeviceInfo',
    index=3,
    containing_service=None,
    input_type=_DEVICEEMPTY,
    output_type=_DEVICEINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DEVICE)

DESCRIPTOR.services_by_name['Device'] = _DEVICE

# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for an PwRPC (Pigweed RPC) locking capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class PwRPCLockBase(capability_base.CapabilityBase):
  """Pigweed RPC locking capability for devices communicating over PwRPC."""

  @abc.abstractmethod
  def lock(self, no_wait: bool = False) -> None:
    """Locks the device.

    Args:
      no_wait: If true, returns before verifying the locked state.
    """

  @abc.abstractmethod
  def unlock(self, no_wait: bool = False) -> None:
    """Unlocks the device.

    Args:
      no_wait: If true, returns before verifying the locked state.
    """

  @property
  @abc.abstractmethod
  def state(self) -> bool:
    """The lock state of the device.

    Returns:
      True if the device is locked, false if it's unlocked.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for the device power capability."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class DevicePowerBase(capability_base.CapabilityBase):
  """Abstract Base class for the device power capability."""

  @abc.abstractproperty
  def hub_name(self):
    """Name of the hub the device is attached to."""

  @abc.abstractproperty
  def hub_type(self):
    """Type of hub the check is attached to."""

  @abc.abstractproperty
  def port_mode(self):
    """port mode."""

  @abc.abstractproperty
  def port_number(self):
    """Port number the device is attached to."""

  @abc.abstractmethod
  def cycle(self, no_wait=False):
    """Power off then power on the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """

  @abc.abstractmethod
  def on(self, no_wait=False):
    """Turn on power to the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """

  @abc.abstractmethod
  def off(self):
    """Turn off power to the device."""



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shell capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base

_DEVICE_CAPABILITY_NAME = "shell_capability"
_SHELL_CMD = u"{cmd};echo Return Code: $?\n"
_SHELL_REGEX = r"(.*)Return Code: (\d+)\n"
_TIMEOUT = 10
_TRIES = 2


class ShellBase(capability_base.CapabilityBase):
  """Abstract base class defining the API for the shell() method."""

  def __init__(self,
               send_and_expect,
               device_name,
               shell_cmd=_SHELL_CMD,
               shell_regex=_SHELL_REGEX,
               tries=_TRIES,
               timeout=_TIMEOUT):
    """Initalize the SSH shell capability.

    Args:
        send_and_expect (method): bound send_and_expect method of the device
          class instance.
        device_name (str): name of the device using this capability.
        shell_cmd (str): return code wrapper around the shell command to
          execute.
        shell_regex (str): shell regex to use. Must contain two capturing
          groups: one for the output and one for the return code.
        tries (int): how many times to try sending the shell command.
        timeout (float): shell timeout in seconds.
    """
    super(ShellBase, self).__init__(device_name=device_name)

    self._send_and_expect = send_and_expect
    self._shell_regex = shell_regex
    self._shell_cmd = shell_cmd
    self._tries = tries
    self._timeout = timeout

  @classmethod
  def get_capability_name(cls):
    """Override the default to avoid a name conflict with the "shell" method."""
    return _DEVICE_CAPABILITY_NAME

  @abc.abstractmethod
  def shell(self,
            command,
            command_name="shell",
            timeout=10,
            port=0,
            include_return_code=False):
    """Sends command and returns response and optionally return code.

    Args:
        command (str): Command to send to the device.
        command_name (str): Identifier for command.
        timeout (float): Time in seconds to wait for device to respond.
        port (int): Which port to send on, 0 or 1.
        include_return_code (bool): flag indicating return code should be
          returned.

    Raises:
        DeviceError: if communication fails.

    Note:
        Can try multiple times as connection can sometimes fail.
        See the init args for setting the number of retry attempts.

    Returns:
        str: If include_return_code is False return the device response to
        the command.
        tuple: If include_return_code is True return the device response and
        return code.
    """

  @abc.abstractmethod
  def has_command(self, binary_name):
    """Returns if binary_name is installed on the device.

    Args:
        binary_name (str): name of the executable.

    Returns:
        bool: True if the executable is found on the device, False otherwise.
    """



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC transport class."""
import fcntl
import queue
import select
import threading
import types
from typing import Any, Collection, Dict, Optional, Tuple
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.switchboard.transports import transport_base
import serial

# TODO(b/181734752): Remove conditional imports of Pigweed
try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  import pw_rpc
  from pw_rpc import callback_client
  from pw_hdlc import rpc
  from pw_hdlc import decode
  from pw_protobuf_compiler import python_protos
  # pytype: enable=import-error
  PIGWEED_IMPORT = True
except ImportError:
  pw_rpc = None
  callback_client = None
  rpc = None
  decode = None
  python_protos = None
  PIGWEED_IMPORT = False

_STDOUT_ADDRESS = 1
_DEFAULT_ADDRESS = ord("R")
_JOIN_TIMEOUT_SEC = 1  # seconds
_SELECT_TIMEOUT_SEC = 0.1  # seconds
logger = gdm_logger.get_logger()


class PwHdlcRpcClient:
  """Pigweed HDLC RPC Client.

  Adapted from https://pigweed.googlesource.com/pigweed/pigweed/+/refs/heads/master/pw_hdlc/py/pw_hdlc/rpc.py. # pylint: disable=line-too-long
  """

  def __init__(self,
               serial_instance: serial.Serial,
               protobufs: Collection[types.ModuleType]):
    """Creates an RPC client configured to communicate using HDLC.

    Args:
      serial_instance: Serial interface instance.
      protobufs: Proto modules.
    """
    if not PIGWEED_IMPORT:
      raise errors.DependencyUnavailableError(
          "Pigweed python packages are not available in this environment.")

    protos = python_protos.Library.from_paths(protobufs)
    client_impl = callback_client.Impl()
    channels = rpc.default_channels(serial_instance.write)
    self._client = pw_rpc.Client.from_modules(client_impl,
                                              channels,
                                              protos.modules())
    self._serial = serial_instance
    self._stop_event = threading.Event()
    self._worker = None
    self.log_queue = queue.Queue()

  def is_alive(self) -> bool:
    """Returns true if the worker thread has started."""
    return self._worker is not None and self._worker.is_alive()

  def start(self) -> None:
    """Creates and starts the worker thread if it hasn't been created."""
    if self._stop_event.is_set():
      self._stop_event.clear()
    if self._worker is None:
      self._worker = threading.Thread(target=self.read_and_process_data)
      self._worker.start()

  def close(self) -> None:
    """Sets the threading event and joins the worker thread."""
    self._stop_event.set()
    if self._worker is not None:
      self._worker.join(timeout=_JOIN_TIMEOUT_SEC)
      if self._worker.is_alive():
        raise errors.DeviceError(
            f"The child thread failed to join after {_JOIN_TIMEOUT_SEC} seconds"
            )
      self._worker = None

  def rpcs(self, channel_id: Optional[int] = None) -> Any:
    """Returns object for accessing services on the specified channel.

    Args:
      channel_id: None or RPC channel id.

    Returns:
      RPC instance over the specificed channel.
    """
    if channel_id is None:
      return next(iter(self._client.channels())).rpcs
    return self._client.channel(channel_id).rpcs

  def read_and_process_data(self) -> None:
    """Continuously reads and handles HDLC frames."""
    decoder = decode.FrameDecoder()
    while not self._stop_event.is_set():
      readfds, _, _ = select.select([self._serial], [], [], _SELECT_TIMEOUT_SEC)
      for fd in readfds:
        try:
          data = fd.read()
        except Exception:  # pylint: disable=broad-except
          logger.exception(
              "Exception occurred when reading in PwHdlcRpcClient thread.")
          data = None
        if data:
          for frame in decoder.process_valid_frames(data):
            self._handle_frame(frame)

  def _handle_rpc_packet(self, frame: Any) -> None:
    """Handler for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    if not self._client.process_packet(frame.data):
      logger.error(f"Packet not handled by RPC client: {frame.data}")

  def _push_to_log_queue(self, frame: Any) -> None:
    """Pushes the HDLC log in frame into the log queue.

    Args:
      frame: HDLC frame packet.
    """
    self.log_queue.put(frame.data + b"\n")

  def _handle_frame(self, frame: Any) -> None:
    """Private method for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    try:
      if not frame.ok():
        return
      if frame.address == _DEFAULT_ADDRESS:
        self._handle_rpc_packet(frame)
      elif frame.address == _STDOUT_ADDRESS:
        self._push_to_log_queue(frame)
      else:
        logger.warning(f"Unhandled frame for address {frame.address}: {frame}")
    except Exception:  # pylint: disable=broad-except
      logger.exception("Exception occurred in PwHdlcRpcClient.")


class PigweedRPCTransport(transport_base.TransportBase):
  """Performs transport communication using the Pigweed RPC to end devices."""

  def __init__(self,
               comms_address: str,
               protobufs: Collection[types.ModuleType],
               baudrate: int,
               auto_reopen: bool = True,
               open_on_start: bool = True):
    super().__init__(
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)
    self.comms_address = comms_address
    self._protobufs = protobufs
    self._serial = serial.Serial()
    self._serial.port = comms_address
    self._serial.baudrate = baudrate
    self._serial.timeout = 0.01
    self._hdlc_client = PwHdlcRpcClient(self._serial, protobufs)

  def is_open(self) -> bool:
    """Returns True if the PwRPC transport is connected to the target.

    Returns:
      True if transport is open, False otherwise.
    """
    return self._serial.isOpen() and self._hdlc_client.is_alive()

  def _close(self) -> None:
    """Closes the PwRPC transport."""
    self._hdlc_client.close()
    fcntl.flock(self._serial.fileno(), fcntl.LOCK_UN)
    self._serial.close()

  def _open(self) -> None:
    """Opens the PwRPC transport."""
    self._serial.open()
    fd = self._serial.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
    self._hdlc_client.start()

  def _read(self, size: int, timeout: float) -> bytes:
    """Returns Pigweed log from the HDLC channel 1.

    Args:
      size: Not used.
      timeout: Maximum seconds to wait to read bytes.

    Returns:
      bytes read from transport or None if no bytes were read
    """
    # Retrieving logs from queue doesn't support size configuration.
    del size
    try:
      return self._hdlc_client.log_queue.get(timeout=timeout)
    except queue.Empty:
      return b""

  def _write(self, data: str, timeout: Optional[float] = None) -> int:
    """Dummy method for Pigweed RPC.

    Declared to pass the inheritance check of TransportBase.

    Args:
      data: Not used.
      timeout: Not used.

    Returns:
      int: Not used.
    """
    return 0

  def rpc(self,
          service_name: str,
          event_name: str,
          **kwargs: Dict[str, Any]) -> Tuple[bool, bytes]:
    """RPC call to the Matter endpoint with given service and event name.

    Args:
      service_name: PwRPC service name.
      event_name: Event name in the given service instance.
      **kwargs: Arguments for the event method.

    Returns:
      (RPC ack value, RPC encoded payload in bytes)
    """
    client_channel = self._hdlc_client.rpcs().chip.rpc
    service = getattr(client_channel, service_name)
    event = getattr(service, event_name)
    ack, payload = event(**kwargs)
    return ack.ok(), payload.SerializeToString()

  def echo_rpc(self, msg: str) -> Tuple[bool, str]:
    """Calls the Echo RPC endpoint.

    Sends a message to the echo endpoint and returns the response back. Uses a
    different namespace (pw.rpc) than the rest of Matter endpoints (chip.rpc).
    Only used by the Pigweed Echo example app:
    https://github.com/project-chip/connectedhomeip/tree/master/examples/pigweed-app/esp32#chip-esp32-pigweed-example-application

    Args:
      msg: Echo message to send.

    Returns:
      (RPC ack value, Echo message)
    """
    client_channel = self._hdlc_client.rpcs().pw.rpc
    echo_service = client_channel.EchoService
    ack, payload = echo_service.Echo(msg=msg)
    return ack.ok(), payload.msg



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Used to identify device output lines as Log, Response, or Unknown.

AllUnknownIdentifier (default) which classifies all lines
from all ports as unknown and accepts them for every line_type provided.

AllLogIdentifier classifies all lines from all ports as log lines and only
accepts them for "all" or "log" line_type values.

AllResponseIdentifier classifies all lines from all ports as response lines
and only accepts them for "all" or "response" line_type values.

PortLogIdentifier classifies lines from a particular port as log lines.
All others are responses.

RegexLogIdentifier classifies lines that match a regex as log lines.
All others are unknown.

RegexResponseIdentifier classifies lines that match a regex as response lines.
All others are unknown.
"""
import abc
import re
from typing import Optional

from typing_extensions import Literal

LINE_TYPE_ALL = "all"
LINE_TYPE_LOG = "log"
LINE_TYPE_RESPONSE = "response"
LINE_TYPE = Literal[LINE_TYPE_ALL, LINE_TYPE_LOG, LINE_TYPE_RESPONSE]


class LineIdentifier(abc.ABC):
  """Base class used to identify device output lines as Log, Response, or Unknown."""

  @abc.abstractmethod
  def accept(self, port, line, line_type):
    """Returns True if line is identified to be part of line_type requested.

    Args:
        port (int): which transport line came from.
        line (str): device output
        line_type (str): line type requested. Options ("all", "log",
          "response")

    Returns:
        bool: whether or not line matches request type.
    """


class AllLogIdentifier(LineIdentifier):
  """Identifies all output as logs."""

  def accept(self, port, line, line_type):
    """Identifies all lines from all ports as log lines."""

    return line_type in [LINE_TYPE_ALL, LINE_TYPE_LOG]


class AllResponseIdentifier(LineIdentifier):
  """Identifies all output as responses."""

  def accept(self, port, line, line_type):
    return line_type in [LINE_TYPE_ALL, LINE_TYPE_RESPONSE]


class AllUnknownIdentifier(LineIdentifier):
  """Identifies all output as unknown."""

  def accept(self, port, line, line_type):
    return True


class PortLogIdentifier(LineIdentifier):
  """Identifies all output from certain ports as logs.

  Note:
      Output from all other ports is considered a reponse.
  """

  def __init__(self, log_ports=None):
    """Identifies lines from log ports as logs.

    Args:
        log_ports (list): ports to identify as logging ports.
    """
    if not log_ports:
      log_ports = [1]
    self.log_ports = log_ports

  def accept(self, port, line, line_type):
    if port in self.log_ports:
      return line_type in [LINE_TYPE_ALL, LINE_TYPE_LOG]
    else:
      return line_type in [LINE_TYPE_ALL, LINE_TYPE_RESPONSE]


class RegexIdentifier(LineIdentifier):
  """Identifies output based on a log or a response regex (or both)."""

  def __init__(self, log_pattern: Optional[str] = None,
               response_pattern: Optional[str] = None) -> None:
    """Initializes a RegexIdentifier.

    Args:
      log_pattern: Regular expression which matches logs.
      response_pattern: Regular expression which matches command responses.
    """
    if log_pattern is None and response_pattern is None:
      raise ValueError("log_pattern or response_pattern must be specified.")
    if log_pattern is not None:
      self._log_pattern = re.compile(log_pattern)
    else:
      self._log_pattern = None
    if response_pattern is not None:
      self._response_pattern = re.compile(response_pattern)
    else:
      self._response_pattern = None

  def accept(self, port: int, line: str, line_type: LINE_TYPE) -> bool:
    """Returns whether the given line should be accepted."""
    if line_type == LINE_TYPE_ALL:
      return True

    if line_type == LINE_TYPE_LOG:
      if self._log_pattern is not None:
        return bool(re.search(self._log_pattern, line))
      else:
        return not bool(re.search(self._response_pattern, line))

    if line_type == LINE_TYPE_RESPONSE:
      if self._response_pattern is not None:
        return bool(re.search(self._response_pattern, line))
      else:
        return not bool(re.search(self._log_pattern, line))

    return True


class RegexLogIdentifier(RegexIdentifier):
  """Identifies all output as logs if regex matches else its Unknown."""

  def __init__(self, log_pattern: str) -> None:
    super().__init__(log_pattern=log_pattern)


class RegexResponseIdentifier(RegexIdentifier):
  """Identifies all output as response if regex matches else its Unknown."""

  def __init__(self, response_pattern: str) -> None:
    super().__init__(response_pattern=response_pattern)


class MultiportIdentifier(LineIdentifier):
  """Generic line identifier for several ports.

  Each port uses a different identifier.
  MultiportIdentifier delegates to the appropriate identifier.
  """

  def __init__(self, identifiers):
    """Initializes the MultiportIdentifier.

    Args:
        identifiers (list): list of LineIdentifier objects.
    """
    self._identifiers = identifiers

  def accept(self, port, line, line_type):
    """Returns whether the given line should be accepted (True) or discarded (False)."""
    if port >= len(self._identifiers):
      return True
    return self._identifiers[port].accept(
        port=port, line=line, line_type=line_type)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the PwRPC (Pigweed RPC) locking capability."""
import enum
from typing import Any, Callable, Dict

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_lock_base
from gazoo_device.protos import locking_service_pb2
from gazoo_device.switchboard.transports import pigweed_rpc_transport


logger = gdm_logger.get_logger()


class LockedState(enum.Enum):
  LOCKED = True
  UNLOCKED = False


class PwRPCLockDefault(pwrpc_lock_base.PwRPCLockBase):
  """Pigweed RPC locking capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               expect_locking_regexes: Dict[bool, str],
               expect_timeout: int,
               switchboard_call: Callable[..., Any],
               switchboard_call_expect: Callable[..., Any]):
    """Initializes an instance of the PwRPCLockDefault capability.

    Args:
      device_name: Device name used for logging.
      expect_locking_regexes: Expected regexes for locking and unlocking, the
      dict format: {LockedState.LOCKED: "<locked regex>", LockedState.UNLOCKED:
        "<unlocked regex>"}.
      expect_timeout: Timeout (s) to wait for the expected regex.
      switchboard_call: The switchboard.call method.
      switchboard_call_expect: The switchboard.call_and_expect method.
    """
    super().__init__(device_name=device_name)
    self._expect_locking_regexes = expect_locking_regexes
    self._expect_timeout = expect_timeout
    self._switchboard_call = switchboard_call
    self._switchboard_call_expect = switchboard_call_expect

  @decorators.CapabilityLogDecorator(logger)
  def lock(self, no_wait: bool = False) -> None:
    """Locks the device.

    Args:
      no_wait: Returns before verifying the locked state if true.
    """
    self._lock_unlock(True, no_wait)

  @decorators.CapabilityLogDecorator(logger)
  def unlock(self, no_wait: bool = False) -> None:
    """Unlocks the device.

    Args:
      no_wait: Returns before verifying the locked state if true.
    """
    self._lock_unlock(False, no_wait)

  @decorators.DynamicProperty
  def state(self) -> bool:
    """The lock state of the device.

    Returns:
      True if the device is locked, false if it's unlocked.

    Raises:
      DeviceError: Ack value is false.
    """
    ack, state_in_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=("Locking", "Get"),
        method_kwargs={})

    if not ack:
      raise errors.DeviceError(
          f"Getting device {self._device_name} locked state failed: "
          "device did not acknowledge the RPC.")

    state = locking_service_pb2.LockingState.FromString(state_in_bytes)
    return state.locked

  def _lock_unlock(self, locked: bool, no_wait: bool = False) -> None:
    """Locks or unlocks the device.

    Args:
      locked: Locks the device if true, unlocks the device if false.
      no_wait: Returns before verifying the locked state if true.

    Raises:
      DeviceError: Ack value is false or the device does not transition to the
      appropriate state.
    """
    regex_type = LockedState.LOCKED if locked else LockedState.UNLOCKED
    expect_regex = self._expect_locking_regexes[regex_type]

    _, (ack, _) = self._switchboard_call_expect(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        pattern_list=[expect_regex],
        timeout=self._expect_timeout,
        method_args=("Locking", "Set"),
        method_kwargs={"locked": locked},
        raise_for_timeout=True)

    action = "Locking" if locked else "Unlocking"
    error_mesg = f"{action} device {self._device_name} failed: "
    if not ack:
      raise errors.DeviceError(
          error_mesg + "device did not acknowledge the RPC.")
    if not no_wait:
      if locked != self.state:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            error_mesg + f"device's locked state remains {self.state}.")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines errors raised by GDM.

The error subclasses are intended to make it easier to distinguish between and
handle different types of error exceptions.

error codes:
    1           Generic catch-all for DeviceError
    10 - 19     TimeoutError exceptions
    30 - ...    CheckDeviceReadyError exceptions
"""
from typing import List

from gazoo_device import _version
from gazoo_device import data_types
from gazoo_device import extensions
from gazoo_device import gdm_logger

logger = gdm_logger.get_logger()


def _get_version_string() -> str:
  """Returns version of GDM and of all registered extension packages."""
  return " GDM Version: {}. Registered extension packages: {}.".format(
      _version.version, extensions.get_registered_package_info())


class DeviceError(Exception):
  """Basic exception for errors raised by devices.

  Attributes:
      err_code (int): numeric code of the error.
  """
  err_code = 1

  def __init__(self, msg):
    """Inits DeviceError with 'msg' (an error message string).

    Args:
        msg (str or Exception): an error message string or an Exception
          instance.

    Note: Additionally, logs 'msg' to debug log level file.
        example: msg = 'Device {} failed to upgrade to target version {}. It
          is now at version {}' DeviceError(msg.format(self.name,
          target_version, new_version)).
    """
    super().__init__(msg)
    logger.debug(repr(self))


class ParserError(DeviceError):
  """Raised when log parser fails."""
  err_code = 5


class PackageRegistrationError(Exception):
  """Raised when GDM fails to register an extension package."""
  err_code = 6

  def __init__(self, msg: str, package_name: str):
    super().__init__("Failed to register package {!r} with GDM architecture. {}"
                     .format(package_name, msg))


class CommunicationTimeoutError(DeviceError):
  """Exceptions raised due to a timeout."""
  err_code = 10


class TransportNotAvailableError(DeviceError):
  """Exception raised when a transport is not available."""
  err_code = 22


class CheckDeviceReadyError(DeviceError):
  """DeviceError variant used in device ready checks."""
  err_code = 30

  def __init__(self,
               device_name,
               msg,
               reason=None,
               details=None,
               recovery=None):
    """Inits a CheckDeviceReadyError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        reason (str): An optional message string describing the reason for
          the error.
        details (str): An optional message string describing error details.
        recovery (str): An optional message string describing further
          recovery options if attempted recover does not work.

    Note: Default suggested recovery is to reflash the device with a valid
      build.
    """
    error_str = "{} {}.".format(device_name, msg)
    if reason:
      error_str += " Reason: {}.".format(reason)
    if details:
      error_str += " Details: {}.".format(details)
    if recovery:
      error_str += " Recovery: {}.".format(recovery)
    error_str += _get_version_string()

    self.properties = {}  # persistent properties of the device
    self.checks_passed = [
    ]  # names of health check methods that passed prior to the failure

    super(CheckDeviceReadyError, self).__init__(error_str)


class DeviceNotConnectedError(CheckDeviceReadyError):
  """Raised when device does not appear to be connected to host computer."""
  err_code = 31


class DeviceNotBootupCompleteError(CheckDeviceReadyError):
  """Raised when device is still booting up."""
  err_code = 32


class DeviceNotResponsiveError(CheckDeviceReadyError):
  """Raised when a command sent to the device times out without receiving a response."""
  err_code = 33

  def __init__(self, device_name, msg, timeout=None, details=None):
    """Inits a DeviceNotResponsiveError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        timeout (int): Number of seconds before timing out while talking to
          device.
        details (str): An optional message string describing error details.
    """
    reason = "timed out "
    if timeout:
      reason += "after {} seconds ".format(timeout)
    reason += "when talking to the DUT"
    super(DeviceNotResponsiveError, self).__init__(
        device_name, msg, reason=reason, details=details)


class NoLogsStreamingError(CheckDeviceReadyError):
  """Raised when expected log output is not seen from the device."""
  err_code = 34

  def __init__(self, device_name, patterns, command=None, timeout=None):
    """Inits a NoLogsStreamingError exception.

    Args:
        device_name (str): The name of the device.
        patterns (list): Pattern list that was expected from the device.
        command (str): Command that was issued in an attempt to find
          streaming logs.
        timeout (int): Number of seconds spent waiting for logs.
    """
    msg = "patterns {!r} not found in logs".format(patterns)
    if command:
      msg += " after issuing {!r} command".format(command)
    reason = "no logs streaming from UART"
    if timeout:
      reason += " in {} seconds".format(timeout)
    super(NoLogsStreamingError, self).__init__(
        device_name, msg=msg, reason=reason)


class NonRootUserError(CheckDeviceReadyError):
  """Raised when the device is logged in as a non-root user."""
  err_code = 35

  def __init__(self, device_name, msg, details=None):
    """Inits a NonRootUserError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
    """
    super(NonRootUserError, self).__init__(
        device_name, msg, reason="logged in as non-root user", details=details)


class LoggingAgentNotEnabledError(CheckDeviceReadyError):
  """Raised when GDM detects that the logging agent is not installed."""
  err_code = 36

  def __init__(self, device_name, msg):
    """Inits a LoggingAgentNotEnabledError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(LoggingAgentNotEnabledError, self).__init__(
        device_name, msg, reason="agent not installed")


class BootloaderModeError(CheckDeviceReadyError):
  """Raised when GDM detects that the device is at the bootloader prompt."""
  err_code = 37

  def __init__(self, device_name, msg):
    """Inits a BootloaderModeError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(BootloaderModeError, self).__init__(
        device_name, msg, reason="found a bootloader prompt")


class DfuModeError(CheckDeviceReadyError):
  """Raised when GDM detects that the device is at the DFU prompt."""
  err_code = 38

  def __init__(self, device_name, msg):
    """Inits a DfuModeError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(DfuModeError, self).__init__(
        device_name, msg, reason="found a DFU Linux prompt")


class DeviceNotLoggedInError(CheckDeviceReadyError):
  """Raised when the device is at a login prompt."""
  err_code = 39

  def __init__(self, device_name, msg, details=None):
    """Inits a DeviceNotLoggedInError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
    """
    super(DeviceNotLoggedInError, self).__init__(
        device_name, msg, reason="device not logged in", details=details)


class UnterminatedQuoteError(CheckDeviceReadyError):
  """Raised when GDM detects a shell continuation prompt."""
  err_code = 40

  def __init__(self, device_name, msg):
    """Inits a UnterminatedQuoteError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(UnterminatedQuoteError, self).__init__(
        device_name, msg, reason="found an 'unterminated quote' prompt")


class StorageNotHealthyError(CheckDeviceReadyError):
  """Raised when GDM detects a problem with device storage."""
  err_code = 41

  def __init__(self, device_name, msg, details=None, recovery=None):
    """Inits a StorageNotHealthyError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
        recovery (str): An optional message string describing further
          recovery options if attempted recover does not work.
    """
    recovery = recovery or "reflash the device with a valid build"
    super(StorageNotHealthyError, self).__init__(
        device_name, msg, details=details, recovery=recovery)


class BatteryTooLowError(CheckDeviceReadyError):
  """Raised when device battery state is too low to function properly."""
  err_code = 42

  def __init__(self, device_name, msg):
    """Inits a BatteryTooLowError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(BatteryTooLowError, self).__init__(
        device_name, msg, reason="battery is too low or dead")


class PasswordNotDisabledError(CheckDeviceReadyError):
  """Raised when the device root user has a password set unexpectedly."""
  err_code = 43

  def __init__(self, device_name, msg):
    """Inits a PasswordNotDisabledError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(PasswordNotDisabledError, self).__init__(
        device_name, msg, reason="root user still has a password")


class FastbootModeError(CheckDeviceReadyError):
  """Raised when GDM detects that the device is in fast-boot mode."""
  err_code = 44

  def __init__(self, device_name, msg, details=None):
    """Inits a FastbootModeError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
    """
    super(FastbootModeError, self).__init__(
        device_name, msg, reason="device is in fastboot mode", details=details)


class DeviceNotRootedError(CheckDeviceReadyError):
  """Raised when GDM detects that the device is not rooted."""
  err_code = 45

  def __init__(self, device_name, msg):
    """Inits a DeviceNotRootedError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(DeviceNotRootedError, self).__init__(
        device_name, msg, reason="device is not rooted")


class DeviceNotADBConnectedError(CheckDeviceReadyError):
  """Raised when the device is not available through ADB."""
  err_code = 46

  def __init__(self, device_name, msg, details=None):
    """Inits a DeviceNotADBConnectedError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
    """
    super(DeviceNotADBConnectedError, self).__init__(
        device_name, msg, reason="device not connected to ADB", details=details)


class DeviceBaudrateMismatchError(CheckDeviceReadyError):
  """Raised when the device baudrate doesn't match the target baudrate."""
  err_code = 47

  def __init__(self, device_name, msg, target_baudrate=None):
    """Inits a DeviceBaudrateMismatchError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        target_baudrate (int): The baudrate that the device should be at.
    """
    if not target_baudrate:
      target_baudrate = "UNKNOWN"
    self.target_baudrate = target_baudrate
    reason = "Target baudrate: {}".format(target_baudrate)
    super(DeviceBaudrateMismatchError, self).__init__(
        device_name, msg, reason=reason)


class HubFailedToRespondError(CheckDeviceReadyError):
  """Raised when Hub device failes to respond."""
  err_code = 48

  def __init__(self, device_name, msg, details=None):
    """Inits a HubFailedToRespondError exception.

    Args:
        device_name (str): The name of the hub device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing error details.
    """
    super(HubFailedToRespondError, self).__init__(
        device_name, msg, reason="hub needs reboot", details=details)


class DeviceMissingPackagesError(CheckDeviceReadyError):
  """Raised when the device is missing some packages required for testing."""
  err_code = 49

  def __init__(self, device_name, msg, package_list=None, details=None):
    """Initializes a DeviceMissingPackagesError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        package_list (list or None): list of missing package names
          (strings).
        details (str): An optional message string describing additional
          error details.
    """
    self.package_list = package_list

    if package_list is None:
      packages = "[<no package list available>]"
    else:
      packages = str(package_list)

    super(DeviceMissingPackagesError, self).__init__(
        device_name,
        msg,
        reason="missing testing packages {}".format(packages),
        details=details)


class DevicePropertiesNotAvailableError(CheckDeviceReadyError):
  """Raised when a device without shell does not have all properties set."""
  err_code = 51

  def __init__(self, device_name, msg, details=None):
    """Inits a DevicePropertiesNotAvailableError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing additional
          error details.
    """
    super(DevicePropertiesNotAvailableError, self).__init__(
        device_name,
        msg,
        reason="not all required properties parsed from logs",
        details=details)


class DeviceBinaryMissingError(CheckDeviceReadyError):
  """Raised when a device does not have a required binary installed."""
  err_code = 52

  def __init__(self, device_name, msg):
    """Inits a DeviceBinaryMissingError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(DeviceBinaryMissingError, self).__init__(
        device_name, msg, reason="Required binary not found on device")


class SwitchboardCreationError(CheckDeviceReadyError):
  """Raised if unable to initiate switchboard."""
  err_code = 53

  def __init__(self, device_name, msg):
    """Inits a SwitchboardCreationError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
    """
    super(SwitchboardCreationError, self).__init__(
        device_name, msg, reason="Unable to create switchboard.")


class DeviceInFactoryTestModeError(CheckDeviceReadyError):
  """Raised when device is in factory test mode."""
  err_code = 54

  def __init__(self, device_name):
    """Inits DeviceInFactoryTestModeError.

    Args:
        device_name (str): The name of the device.
    """
    super(DeviceInFactoryTestModeError,
          self).__init__(device_name, "Device in factory test mode")


class DeviceIpRouteError(CheckDeviceReadyError):
  """Raised when device defaults to IP routing over the Ethernet interface instead of Wi-Fi."""
  err_code = 55

  def __init__(self,
               device_name,
               msg=None,
               reason=None,
               details=None,
               recovery=None):
    if msg is None:
      msg = ("defaults to routing IP packets over the Ethernet interface "
             "instead of Wi-Fi")

    super(DeviceIpRouteError, self).__init__(
        device_name, msg=msg, reason=reason, details=details, recovery=recovery)


class DeviceServiceNotYetStartedError(CheckDeviceReadyError):
  """Raised when device services are still getting ready."""
  err_code = 57


class CheckTestbedReadyError(DeviceError):
  """DeviceError variant used in testbed readiness checks."""
  err_code = 58

  def __init__(self, msg, details=None, recovery=None):
    """Inits a CheckTestbedReadyError exception.

    Args:
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing error details.
        recovery (str): An optional message string describing further
          recovery options if attempted recover does not work.
    """
    error_str = "testbed {}.".format(msg)
    if details:
      error_str += " Details: {}.".format(details)
    if recovery:
      error_str += " Recovery: {}.".format(recovery)
    error_str += _get_version_string()

    self.properties = {}  # testing properties of the testbed
    self.checks_passed = [
    ]  # names of health check methods that passed prior to the failure

    super(CheckTestbedReadyError, self).__init__(error_str)


class TestbedWifiNotAvailableError(CheckTestbedReadyError):
  """Raised when a the SSID of a wifi network is not available to the testbed."""
  err_code = 59

  def __init__(self, ssid, available_networks, details=None, recovery=None):
    """Inits a TestbedWifiNotAvailableError exception.

    Args:
        ssid (str): SSID of wifi network.
        available_networks (list): List of available wifi network SSIDs.
        details (str): An optional message string describing error details.
        recovery (str): An optional message string describing further
          recovery options if attempted recover does not work.
    """
    msg = "unable to find available wifi network with ssid {!r}.".format(ssid)
    msg += " Available networks: {}".format(available_networks or "none")
    super(TestbedWifiNotAvailableError, self).__init__(msg, details, recovery)


class ProcessNotRunningError(CheckDeviceReadyError):
  """Raised when an expected process is not running on the device."""
  err_code = 60

  def __init__(self, device_name, msg, details=None):
    """Inits a ProcessNotRunningError exception.

    Args:
        device_name (str): The name of the device.
        msg (str): An error message string of the form <error_message>
          <details>.
        details (str): An optional message string describing error details.
    """
    super(ProcessNotRunningError, self).__init__(
        device_name, msg, details=details)


class IncompatibleFirmwareError(CheckDeviceReadyError):
  """Raised when the device has a firmware build incompatible with GDM."""
  err_code = 63


class DownloadKeyError(DeviceError):
  """Raised when the host is unable to download a key required by GDM."""
  err_code = 64

  def __init__(self,
               key_info: data_types.KeyInfo,
               download_errors: List[Exception]):
    """Initializes a DownloadKeyError exception.

    Args:
        key_info: Key information.
        download_errors: Errors that occurred when attempting to download the
          key.
    """
    self.key_info = key_info
    super().__init__("unable to download {!r} key to host: {}".format(
        key_info, ", ".join(repr(err) for err in download_errors)))


class TestbedMissingExternalStorageError(CheckTestbedReadyError):
  """Raised when devices in the testbed are missing external storage."""
  err_code = 65

  def __init__(self, device_names, details=None, recovery=None):
    """Initializes a TestbedMissingExternalStorageError exception.

    Args:
        device_names (list): names of devices missing external storage.
        details (str): An optional message string describing error details.
        recovery (str): An optional message string describing further
          recovery options if attempted recovery does not work.
    """
    msg = "devices missing external storage: {}".format(device_names)
    super().__init__(msg, details, recovery)


class CapabilityNotReadyError(CheckDeviceReadyError):
  """Raised when capability not ready."""
  err_code = 66


class DependencyUnavailableError(DeviceError):
  """Exception raised when a dependency is not available."""
  err_code = 67


class TestbedFailedToEnsureBuildsError(CheckTestbedReadyError):
  """Raised when upgrade to expected build fails in a testbed."""
  err_code = 68

  def __init__(self, device_upgrade_exceptions):
    msg = "upgrade failed on devices: {}".format(
        ", ".join(device_upgrade_exceptions.keys()))
    details = ". ".join([
        f"{device}: {repr(exception)}"
        for device, exception in device_upgrade_exceptions.items()
    ])
    recovery = "reflash devices"
    super().__init__(msg, details, recovery)
